# model_mlp.py
import ast
from collections import OrderedDict

import numpy as np
import pandas as pd
import torch
from sklearn.metrics import recall_score

import nkululeko.glob_conf as glob_conf
from nkululeko.losses.loss_softf1loss import SoftF1Loss
from nkululeko.models.model import Model
from nkululeko.reporting.reporter import Reporter


class MLPModel(Model):
    """MLP = multi layer perceptron."""

    is_classifier = True

    def __init__(self, df_train, df_test, feats_train, feats_test):
        """Constructor, taking all dataframes.

        Args:
            df_train (pd.DataFrame): The train labels.
            df_test (pd.DataFrame): The test labels.
            feats_train (pd.DataFrame): The train features.
            feats_test (pd.DataFrame): The test features.
        """
        super().__init__(df_train, df_test, feats_train, feats_test)
        super().set_model_type("ann")
        self.name = "mlp"
        self.target = glob_conf.config["DATA"]["target"]
        labels = glob_conf.labels
        self.class_num = len(labels)
        # set up loss criterion
        criterion = self.util.config_val("MODEL", "loss", "cross")
        if criterion == "cross":
            self.criterion = torch.nn.CrossEntropyLoss()
        elif criterion == "f1":
            self.criterion = SoftF1Loss(
                num_classes=self.class_num, weight=None, epsilon=1e-7
            )
        else:
            self.util.error(f"unknown loss function: {criterion}")
        self.util.debug("using model with cross entropy loss function")
        # set up the model, use GPU if availabe
        cuda = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = self.util.config_val("MODEL", "device", cuda)
        try:
            layers_string = glob_conf.config["MODEL"]["layers"]
        except KeyError as ke:
            self.util.error(f"Please provide MODEL layers: {ke}")
        self.util.debug(f"using layers {layers_string}")
        layers = ast.literal_eval(layers_string)
        # with dropout?
        drop = self.util.config_val("MODEL", "drop", False)
        if drop:
            self.util.debug(f"init: training with dropout: {drop}")
        self.model = self.MLP(feats_train.shape[1], layers, self.class_num, drop).to(
            self.device
        )
        self.learning_rate = float(
            self.util.config_val("MODEL", "learning_rate", 0.0001)
        )
        # set up regularization
        self.optimizer = torch.optim.Adam(
            self.model.parameters(), lr=self.learning_rate
        )
        # batch size
        self.batch_size = int(self.util.config_val("MODEL", "batch_size", 8))
        # number of parallel processes
        self.num_workers = self.n_jobs
        if feats_train.isna().to_numpy().any():
            self.util.debug(
                f"Model, train: replacing {feats_train.isna().sum().sum()} NANs"
                " with 0"
            )
            feats_train = feats_train.fillna(0)
        if feats_test.isna().to_numpy().any():
            self.util.debug(
                f"Model, test: replacing {feats_test.isna().sum().sum()} NANs" " with 0"
            )
            feats_test = feats_test.fillna(0)
        # set up the data_loaders
        self.trainloader = self.get_loader(feats_train, df_train, True)
        self.testloader = self.get_loader(feats_test, df_test, False)

    def set_testdata(self, data_df, feats_df):
        self.testloader = self.get_loader(feats_df, data_df, False)

    def reset_test(self, df_test, feats_test):
        self.testloader = self.get_loader(feats_test, df_test, False)

    def train(self):
        self.model.train()
        losses = []
        for features, labels in self.trainloader:
            logits = self.model(features.to(self.device))
            loss = self.criterion(logits, labels.to(self.device, dtype=torch.int64))
            losses.append(loss.item())
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
        self.loss = (np.asarray(losses)).mean()

    def evaluate(self, model, loader, device):
        logits = torch.zeros(len(loader.dataset), self.class_num)
        targets = torch.zeros(len(loader.dataset))
        model.eval()
        losses = []
        with torch.no_grad():
            for index, (features, labels) in enumerate(loader):
                start_index = index * loader.batch_size
                end_index = (index + 1) * loader.batch_size
                if end_index > len(loader.dataset):
                    end_index = len(loader.dataset)
                logits[start_index:end_index, :] = model(features.to(device))
                targets[start_index:end_index] = labels
                loss = self.criterion(
                    logits[start_index:end_index, :].to(device),
                    labels.to(device, dtype=torch.int64),
                )
                losses.append(loss.item())

        self.loss_eval = (np.asarray(losses)).mean()
        predictions = logits.argmax(dim=1)
        uar = recall_score(targets.numpy(), predictions.numpy(), average="macro")
        return uar, targets, predictions, logits

    def get_probas(self, logits):
        # make a dataframe for probabilites (logits)
        proba_d = {}
        classes = self.df_test[self.target].unique()
        classes.sort()
        for c in classes:
            proba_d[c] = []
        for i, c in enumerate(classes):
            proba_d[c] = list(logits.numpy().T[i])
        probas = pd.DataFrame(proba_d)
        probas = probas.set_index(self.df_test.index)
        return probas

    def predict(self):
        _, truths, predictions, logits = self.evaluate(
            self.model, self.testloader, self.device
        )
        uar, _, _, _ = self.evaluate(self.model, self.trainloader, self.device)
        probas = self.get_probas(logits)
        report = Reporter(truths, predictions, self.run, self.epoch, probas=probas)
        try:
            report.result.loss = self.loss
        except AttributeError:  # if the model was loaded from disk the loss is unknown
            pass
        try:
            report.result.loss_eval = self.loss_eval
        except AttributeError:  # if the model was loaded from disk the loss is unknown
            pass
        report.result.train = uar
        return report

    def get_predictions(self):
        _, _, predictions, _ = self.evaluate(self.model, self.testloader, self.device)
        return predictions.numpy()

    def get_loader(self, df_x, df_y, shuffle):
        data = []
        for i in range(len(df_x)):
            data.append([df_x.values[i], df_y[self.target].iloc[i]])
        return torch.utils.data.DataLoader(
            data, shuffle=shuffle, batch_size=self.batch_size
        )

    class MLP(torch.nn.Module):
        def __init__(self, i, layers, o, drop):
            super().__init__()
            sorted_layers = sorted(layers.items(), key=lambda x: x[1])
            layers = OrderedDict()
            layers["0"] = torch.nn.Linear(i, sorted_layers[0][1])
            layers["0_r"] = torch.nn.ReLU()
            for i in range(0, len(sorted_layers) - 1):
                layers[str(i + 1)] = torch.nn.Linear(
                    sorted_layers[i][1], sorted_layers[i + 1][1]
                )
                if drop:
                    layers[str(i) + "_d"] = torch.nn.Dropout(float(drop))
                layers[str(i) + "_r"] = torch.nn.ReLU()
            layers[str(len(sorted_layers) + 1)] = torch.nn.Linear(
                sorted_layers[-1][1], o
            )
            self.linear = torch.nn.Sequential(layers)

        def forward(self, x):
            # x: (batch_size, channels, samples)
            x = x.squeeze(dim=1).float()
            return self.linear(x)

    def predict_shap(self, features):
        # predict outputs for all samples in SHAP format (pd. dataframe)
        results = []
        for index, row in features.iterrows():
            feats = row.values
            res_dict = self.predict_sample(feats)
            class_key = max(res_dict, key=res_dict.get)
            results.append(class_key)
        return results

    def predict_sample(self, features):
        """Predict one sample."""
        with torch.no_grad():
            features = torch.from_numpy(features)
            features = np.reshape(features, (-1, 1)).T
            logits = self.model(features.to(self.device))
            # logits = self.model(features)
        # if tensor conver to cpu
        if isinstance(logits, torch.Tensor):
            logits = logits.cpu()
        a = logits.numpy()
        res = {}
        for i in range(len(a[0])):
            res[i] = a[0][i]
        return res

    def store(self):
        torch.save(self.model.state_dict(), self.store_path)

    def store_as_onnx(self):
        pass

    def load(self, run, epoch):
        self.set_id(run, epoch)
        dir = self.util.get_path("model_dir")
        # name = f'{self.util.get_exp_name()}_{run}_{epoch:03d}.model'
        name = f"{self.util.get_exp_name(only_train=True)}_{self.run}_{self.epoch:03d}.model"
        cuda = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = self.util.config_val("MODEL", "device", cuda)
        layers = ast.literal_eval(glob_conf.config["MODEL"]["layers"])
        self.store_path = dir + name
        drop = self.util.config_val("MODEL", "drop", False)
        if drop:
            self.util.debug(f"loading: dropout set to: {drop}")
        self.model = self.MLP(
            self.feats_train.shape[1], layers, self.class_num, drop
        ).to(self.device)
        self.model.load_state_dict(torch.load(self.store_path))
        self.model.eval()

    def load_path(self, path, run, epoch):
        self.set_id(run, epoch)
        with open(path, "rb") as handle:
            cuda = "cuda" if torch.cuda.is_available() else "cpu"
            self.device = self.util.config_val("MODEL", "device", cuda)
            layers = ast.literal_eval(glob_conf.config["MODEL"]["layers"])
            self.store_path = path
            drop = self.util.config_val("MODEL", "drop", False)
            if drop:
                self.util.debug(f"dropout set to: {drop}")
            self.model = self.MLP(
                self.feats_train.shape[1], layers, self.class_num, drop
            ).to(self.device)
            self.model.load_state_dict(torch.load(self.store_path))
            self.model.eval()



# model_tree.py

from sklearn.tree import DecisionTreeClassifier

from nkululeko.models.model import Model


class Tree_model(Model):
    """An Tree model"""

    is_classifier = True

    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        self.name = "tree"
        self.clf = DecisionTreeClassifier(random_state=42)  # set up the classifier



# nkululeko.py
# Entry script to do a Nkululeko experiment
import argparse
import configparser
from pathlib import Path

import numpy as np

import nkululeko.experiment as exp
from nkululeko.constants import VERSION
from nkululeko.utils.util import Util


def doit(config_file):
    # test if the configuration file exists
    if not Path(config_file).is_file():
        print(f"ERROR: no such file: {config_file}")
        exit()

    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)

    # create a new experiment
    expr = exp.Experiment(config)
    module = "nkululeko"
    expr.set_module(module)
    util = Util(module)
    util.debug(
        f"running {expr.name} from config {config_file}, nkululeko version"
        f" {VERSION}"
    )

    if util.config_val("EXP", "no_warnings", False):
        import warnings

        warnings.filterwarnings("ignore")

    # load the data
    expr.load_datasets()

    # split into train and test
    expr.fill_train_and_tests()
    util.debug(f"train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}")

    # extract features
    expr.extract_feats()

    # initialize a run manager
    expr.init_runmanager()

    # run the experiment
    reports, last_epochs = expr.run()
    result = expr.get_best_report(reports).result.test
    expr.store_report()
    print("DONE")
    return result, int(np.asarray(last_epochs).min())


def main():
    cwd = Path(__file__).parent.absolute()
    parser = argparse.ArgumentParser(description="Call the nkululeko framework.")
    parser.add_argument("--version", action="version", version=f"Nkululeko {VERSION}")
    parser.add_argument("--config", default="exp.ini", help="The base configuration")
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config
    else:
        config_file = cwd / "exp.ini"
    doit(config_file)


if __name__ == "__main__":
    main()  # use this if you want to state the config file path on command line



# process_database --> BAVED
# modified from https://github.com/Strong-AI-Lab/emotion/blob/master/datasets/BAVED/process.py

import argparse
from html import parser
from pathlib import Path

import pandas as pd

emotion_level = ["low", "normal", "high"]
word_map = [
    "اعجبني",
    "لم يعجبني",
    "هذا",
    "الفيلم",
    "رائع",
    "مقول",
    "سيئ",
]

gender_map = {
    "M": "male",
    "F": "female"
}


def main():  
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str, default='BAVED')
    parser.add_argument('--output_dir', type=str, default='.')
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    # read data
    paths = list(data_dir.glob('?/*.wav'))
    emo_levels = [emotion_level[int(p.stem.split('-')[4])] for p in paths]
    speakers = [p.stem.split('-')[0] for p in paths]
    genders = [gender_map[p.stem.split('-')[1].upper()] for p in paths]
    ages = [int(p.stem.split('-')[2]) for p in paths]
    words = [word_map[int(p.stem.split('-')[3])] for p in paths]
    languages = ["arabic" for p in paths]

    # convert to dataframe
    df = pd.DataFrame({"file": paths, "emotion_level": emo_levels,
                      "speakers": speakers, "gender": genders, "age": ages,
                       "word": words, "language": languages})

    # allocate speaker 50, 46, 102 , 55, 51
    test_speakers = ['46', '4', '54', '47', '51']
    df_test = df[df['speakers'].isin(test_speakers)]
    df_train = df.drop(df_test.index)

    # save to CSV
    df.to_csv(output_dir / 'baved.csv', index=False)
    df_train.to_csv(output_dir / 'baved_train.csv', index=False)
    df_test.to_csv(output_dir / 'baved_test.csv', index=False)

    print(f"BAVED: {len(df)} samples, {len(df_train)} train, {len(df_test)} test")

if __name__ == '__main__':
    main()



"""
Segment audio files in a given directory into smaller chunks with a specified length and overlap. This requires the PyDub library to be installed.

Args:
    input_dir (str): The directory containing the audio files to be segmented.
    output_dir (str): The directory where the segmented audio files will be saved.
    segment_length (int): The length of each audio segment in milliseconds.
    overlap (int): The overlap between adjacent audio segments in milliseconds.

This function will recursively search the input directory for all .wav audio files, and then segment each file into smaller chunks with the specified length and overlap. The segmented audio files will be saved in the output directory, preserving the relative directory structure from the input directory.
"""

import argparse
import glob
from pathlib import Path

from pydub import AudioSegment


# list audio files given a directory
def segment_audio(input_dir, output_dir, segment_length, overlap):
    # check if input dir exist
    if not Path(input_dir).exists():
        print(f"Directory {input_dir} does not exist.")
        return

    # check if output dir exist, create if not
    if not Path(output_dir).exists():
        Path(output_dir).mkdir(parents=True)

    audio_files = glob.glob(str(Path(input_dir) / "**" / "*.wav"), recursive=True)

    for audio_file in audio_files:
        # segment into 2 seconds with 1 second overlap (default values)
        audio = AudioSegment.from_file(audio_file)

        segments = []

        for i in range(0, len(audio), segment_length - overlap):
            segment = audio[i : i + segment_length]
            segments.append(segment)

        # Path(output_dir).mkdir(exist_ok=True)
        for i, segment in enumerate(segments):
            # get relative path from input_dir
            relative_path = Path(audio_file).relative_to(input_dir)
            # make output directory if not exist
            output_subdir = Path(output_dir) / relative_path.parent
            output_subdir.mkdir(parents=True, exist_ok=True)
            segment.export(
                str(output_subdir / f"{Path(audio_file).stem}_{i}.wav"),
                format="wav",
            )

    print("DONE.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--input_dir", type=str, default="./")
    # add argument for output_dir
    parser.add_argument(
        "--output_dir",
        type=str,
        default="./segmented_data/",
    )
    parser.add_argument(
        "--segment_length",
        type=int,
        default=2000,
    )
    parser.add_argument(
        "--overlap",
        type=int,
        default=1000,
    )
    args = parser.parse_args()
    segment_audio(
        args.input_dir,
        args.output_dir,
        segment_length=args.segment_length,
        overlap=args.overlap,
    )



"""
Explore the feature sets of a machine learning experiment.

This script is the entry point for the 'explore' module of the nkululeko framework.
It handles loading the experiment configuration, setting up the experiment, and
running various feature exploration techniques based on the configuration.

The script supports the following configuration options:
- `no_warnings`: If set to `True`, it will ignore all warnings during the exploration.
- `feature_distributions`: If set to `True`, it will generate plots of the feature distributions.
- `tsne`: If set to `True`, it will generate a t-SNE plot of the feature space.
- `scatter`: If set to `True`, it will generate a scatter plot of the feature space.
- `spotlight`: If set to `True`, it will generate a 'spotlight' plot of the feature space.
- `shap`: If set to `True`, it will generate SHAP feature importance plots.
- `model`: The type of model to use for the feature exploration (e.g. 'SVM').
- `plot_tree`: If set to `True`, it will generate a decision tree plot.

The script can be run from the command line with the `--config` argument to specify
the configuration file to use. If no configuration file is provided, it will look
for an `exp.ini` file in the same directory as the script.
"""

# explore.py
# explore the feature sets

import argparse
import configparser
from pathlib import Path

from nkululeko.constants import VERSION
from nkululeko.experiment import Experiment
from nkululeko.utils.util import Util


def main():
    parser = argparse.ArgumentParser(
        description="Call the nkululeko EXPLORE framework."
    )
    parser.add_argument("--config", default="exp.ini", help="The base configuration")
    args = parser.parse_args()
    config_file = args.config if args.config is not None else "exp.ini"

    if not Path(config_file).is_file():
        print(f"ERROR: no such file: {config_file}")
        exit()

    config = configparser.ConfigParser()
    config.read(config_file)
    expr = Experiment(config)
    module = "explore"
    expr.set_module(module)
    util = Util(module)
    util.debug(
        f"running {expr.name} from config {config_file}, nkululeko version"
        f" {VERSION}"
    )

    if util.config_val("EXP", "no_warnings", False):
        import warnings

        warnings.filterwarnings("ignore")
    needs_feats = False
    try:
        # load the experiment
        expr.load(f"{util.get_save_name()}")
        needs_feats = True
    except FileNotFoundError:
        # first time: load the data
        expr.load_datasets()

        # split into train and test
        expr.fill_train_and_tests()
        util.debug(
            f"train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}"
        )

        plot_feats = eval(util.config_val("EXPL", "feature_distributions", "False"))
        tsne = eval(util.config_val("EXPL", "tsne", "False"))
        scatter = eval(util.config_val("EXPL", "scatter", "False"))
        spotlight = eval(util.config_val("EXPL", "spotlight", "False"))
        shap = eval(util.config_val("EXPL", "shap", "False"))
        model_type = util.config_val("EXPL", "model", False)
        plot_tree = eval(util.config_val("EXPL", "plot_tree", "False"))
        needs_feats = False
        if plot_feats or tsne or scatter or model_type or plot_tree or shap:
            # these investigations need features to explore
            expr.extract_feats()
            needs_feats = True
            # explore
            expr.init_runmanager()
            expr.runmgr.do_runs()
    expr.analyse_features(needs_feats)
    expr.store_report()
    print("DONE")


if __name__ == "__main__":
    main()



import argparse
import configparser
import os
import os.path
import sys

from nkululeko.nkululeko import doit as nkulu
from nkululeko.test import do_it as test_mod


def doit(cla):
    parser = argparse.ArgumentParser(description="Call the nkululeko framework.")
    parser.add_argument("--config", help="The base configuration")
    parser.add_argument("--mod", default="nkulu", help="Which nkululeko module to call")
    parser.add_argument("--data", help="The databases", nargs="*", action="append")
    parser.add_argument(
        "--label", nargs="*", help="The labels for the target", action="append"
    )
    parser.add_argument(
        "--tuning_params", nargs="*", help="parameters to be tuned", action="append"
    )
    parser.add_argument(
        "--layers",
        nargs="*",
        help="layer config for mlp, e.g. l1:128 ",
        action="append",
    )
    parser.add_argument("--model", default="xgb", help="The model type")
    parser.add_argument("--feat", default="['os']", help="The feature type")
    parser.add_argument("--set", help="The opensmile set")
    parser.add_argument("--target", help="The target designation")
    parser.add_argument("--epochs", help="The number of epochs")
    parser.add_argument("--runs", help="The number of runs")
    parser.add_argument("--learning_rate", help="The learning rate")
    parser.add_argument("--drop", help="The dropout rate [0:1]")

    args = parser.parse_args(cla)

    if args.config is not None:
        config_file = args.config
    else:
        print("ERROR: need config file")
        quit(-1)

    if args.mod is not None:
        nkulu_mod = args.mod

    # test if config is there
    if not os.path.isfile(config_file):
        print(f"ERROR: no such file {config_file}")

    config = configparser.ConfigParser()
    config.read(config_file)
    # fill the config

    if args.data is not None:
        databases = []
        for t in args.data:
            databases.append(t[0])
        print(f"got databases: {databases}")
        config["DATA"]["databases"] = str(databases)
    if args.label is not None:
        labels = []
        for label in args.label:
            labels.append(label[0])
        print(f"got labels: {labels}")
        config["DATA"]["labels"] = str(labels)
    if args.tuning_params is not None:
        tuning_params = []
        for tp in args.tuning_params:
            tuning_params.append(tp[0])
        config["MODEL"]["tuning_params"] = str(tuning_params)
    if args.layers is not None:
        config["MODEL"]["layers"] = args.layers[0][0]
    if args.target is not None:
        config["DATA"]["target"] = args.target
    if args.epochs is not None:
        config["EXP"]["epochs"] = args.epochs
    if args.runs is not None:
        config["EXP"]["runs"] = args.runs
    if args.learning_rate is not None:
        config["MODEL"]["learning_rate"] = args.learning_rate
    if args.drop is not None:
        config["MODEL"]["drop"] = args.drop
    if args.model is not None:
        config["MODEL"]["type"] = args.model
    if args.feat is not None:
        config["FEATS"]["type"] = f"['{args.feat}']"
    if args.set is not None:
        config["FEATS"]["set"] = args.set
    tmp_config = "tmp.ini"
    with open(tmp_config, "w") as tmp_file:
        config.write(tmp_file)

    result, last_epoch = 0, 0
    if nkulu_mod == "nkulu":
        result, last_epoch = nkulu(tmp_config)
    elif nkulu_mod == "test":
        result, last_epoch = test_mod(tmp_config, "test_results.csv")
    else:
        print(f"ERROR: unknown module: {nkulu_mod}, should be [nkulu | test]")
    return result, last_epoch


if __name__ == "__main__":
    cla = sys.argv
    cla.pop(0)
    doit(cla)  # sys.argv[1])



# process_database.py --> Thorsten-Emotional

import argparse
import os
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=Path,
        default=Path('./thorsten-emotional_v02'))
    parser.add_argument('--output_dir', type=Path, default=Path('./'))
    args = parser.parse_args()

    # Read the database
    input_dir = args.data_dir
    output_dir = args.output_dir

    # list of all the WAV files in the database
    wav_list = list(input_dir.glob('**/*.wav'))
    
    file = [p for p in wav_list]
    emotion = [p.parts[1] for p in file]


    # convert to df
    df = pd.DataFrame(data={'file': file, 'emotion': emotion})

    # split to train, dev, test sets using scikit-learn
    train, test = train_test_split(df, test_size=0.2, 
        random_state=42, stratify=df['emotion'])

    # save to csv
    train.to_csv(output_dir / 'thorsten-emotional_train.csv', index=False)
    test.to_csv(output_dir / 'thorsten-emotional_test.csv', index=False)


if __name__ == '__main__':
    main()



# xgrmodel.py

from xgboost.sklearn import XGBRegressor

from nkululeko.models.model import Model


class XGR_model(Model):
    """An XGBoost regression model"""

    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        self.name = "xgr"
        self.is_classifier = False
        self.clf = XGBRegressor()  # set up the regressor



# augmenter.py
import os

import audeer
import audiofile
import pandas as pd
from audiomentations import *
from tqdm import tqdm

from nkululeko.utils.util import Util


class Augmenter:
    """
    augmenting the train split
    """

    def __init__(self, df):
        self.df = df
        self.util = Util("augmenter")
        # Define a standard transformation that randomly add augmentations to files
        # self.audioment = Compose(
        #     [
        #         AddGaussianNoise(min_amplitude=0.001, max_amplitude=0.015, p=0.5),
        #         TimeStretch(min_rate=0.8, max_rate=1.25, p=0.5),
        #         PitchShift(min_semitones=-4, max_semitones=4, p=0.5),
        #         Shift(p=0.5),
        #     ]
        # )
        defaults = 'Compose([AddGaussianNoise(min_amplitude=0.001, max_amplitude=0.05),Shift(p=0.5),BandPassFilter(min_center_freq=100.0, max_center_freq=6000),Limiter(min_threshold_db=-16.0,max_threshold_db=-6.0,threshold_mode="relative_to_signal_peak"),ClippingDistortion(),])'
        audiomentations = self.util.config_val("AUGMENT", "augmentations", defaults)
        self.audioment = eval(audiomentations)

    def changepath(self, fp, np):
        #        parent = os.path.dirname(fp).split('/')[-1]
        fullpath = os.path.dirname(fp)
        #       newpath = f'{np}{parent}'
        #       audeer.mkdir(newpath)
        return fp.replace(fullpath, np)

    def augment(self, sample_selection):
        """
        augment the training files and return a dataframe with new files index.
        """
        files = self.df.index.get_level_values(0).values
        store = self.util.get_path("store")
        filepath = f"{store}augmentations/"
        audeer.mkdir(filepath)
        self.util.debug(f"augmenting {sample_selection} samples to {filepath}")
        newpath = ""
        index_map = {}
        for i, f in enumerate(tqdm(files)):
            signal, sr = audiofile.read(f)
            filename = os.path.basename(f)
            parent = os.path.dirname(f).split("/")[-1]
            sig_aug = self.audioment(samples=signal, sample_rate=sr)
            newpath = f"{filepath}/{parent}/"
            audeer.mkdir(newpath)
            new_full_name = newpath + filename
            audiofile.write(new_full_name, signal=sig_aug, sampling_rate=sr)
            index_map[f] = new_full_name
        df_ret = self.df.copy()

        file_index = df_ret.index.to_series().map(lambda x: index_map[x[0]]).values
        # workaround because i just couldn't get this easier...
        arrays = [
            file_index,
            list(df_ret.index.get_level_values(1)),
            list(df_ret.index.get_level_values(2)),
        ]
        new_index = pd.MultiIndex.from_arrays(arrays, names=("file", "start", "end"))
        df_ret = df_ret.set_index(new_index)

        return df_ret



"""
Ensemble predictions from multiple experiments.

Args:
    config_files (list): List of configuration file paths.
    method (str): Ensemble method to use. Options are 'majority_voting', 'mean', 'max', 'sum', 'uncertainty', 'uncertainty_weighted', 'confidence_weighted', or 'performance_weighted'.
    threshold (float): Threshold for the 'uncertainty' ensemble method (default: 1.0, i.e. no threshold).
    weights (list): Weights for the 'performance_weighted' ensemble method.
    no_labels (bool): Flag indicating whether the predictions have labels or not.

Returns:
    pandas.DataFrame: The ensemble predictions.

Raises:
    ValueError: If an unknown ensemble method is provided.
    AssertionError: If the number of config files is less than 2 for majority voting.
"""

#!/usr/bin/env python
# -*- coding: utf-8 -*-


import configparser
import time
from argparse import ArgumentParser
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
from sklearn.metrics import balanced_accuracy_score, classification_report

from nkululeko.constants import VERSION
from nkululeko.experiment import Experiment
from nkululeko.utils.util import Util

# import torch

# Constants
DEFAULT_METHOD = "mean"
DEFAULT_OUTFILE = "ensemble_result.csv"


def majority_voting(ensemble_preds_ls):
    all_predictions = pd.concat([df["predicted"] for df in ensemble_preds_ls], axis=1)
    return all_predictions.mode(axis=1).iloc[:, 0]


def mean_ensemble(ensemble_preds, labels):
    for label in labels:
        ensemble_preds[label] = ensemble_preds[label].mean(axis=1)
    return ensemble_preds[labels].idxmax(axis=1)


def max_ensemble(ensemble_preds, labels):
    for label in labels:
        ensemble_preds[label] = ensemble_preds[label].max(axis=1)
    return ensemble_preds[labels].idxmax(axis=1)


def sum_ensemble(ensemble_preds, labels):
    for label in labels:
        ensemble_preds[label] = ensemble_preds[label].sum(axis=1)
    return ensemble_preds[labels].idxmax(axis=1)


def uncertainty_ensemble(ensemble_preds_ls, labels, threshold):
    final_predictions = []
    final_uncertainties = []

    for idx in ensemble_preds_ls[0].index:
        uncertainties = [df.loc[idx, "uncertainty"] for df in ensemble_preds_ls]
        min_uncertainty_idx = np.argmin(uncertainties)
        min_uncertainty = uncertainties[min_uncertainty_idx]

        if min_uncertainty <= threshold:
            # Use the prediction with low uncertainty
            final_predictions.append(
                ensemble_preds_ls[min_uncertainty_idx].loc[idx, "predicted"]
            )
            final_uncertainties.append(min_uncertainty)
        else:  # for uncertainty above threshold
            # Calculate mean of probabilities same class different model
            mean_probs = np.mean(
                [df.loc[idx, labels].values for df in ensemble_preds_ls], axis=0
            )
            final_predictions.append(labels[np.argmax(mean_probs)])
            final_uncertainties.append(np.mean(uncertainties))

    return final_predictions


def uncertainty_weighted_ensemble(ensemble_preds_ls, labels):
    """Weighted ensemble based on uncertainty, normalized for each class"""
    final_predictions = []
    final_uncertainties = []

    for idx in ensemble_preds_ls[0].index:
        uncertainties = [df.loc[idx, "uncertainty"] for df in ensemble_preds_ls]
        # Convert uncertainties to accuracies/confidence
        accuracies = [1 - uncertainty for uncertainty in uncertainties]

        # Calculate weights (inverse of uncertainties)
        weights = [
            1 / uncertainty if uncertainty != 0 else 1e10
            for uncertainty in uncertainties
        ]

        # Normalize weights for each class
        total_weight = sum(weights)
        normalized_weights = [w / total_weight for w in weights]

        # Calculate weighted probabilities for each class
        weighted_probs = {label: 0 for label in labels}
        for df, weight in zip(ensemble_preds_ls, normalized_weights):
            for label in labels:
                weighted_probs[label] += df.loc[idx, label] * weight

        # Select the class with the highest weighted probability
        predicted_class = max(weighted_probs, key=weighted_probs.get)
        final_predictions.append(predicted_class)

        # Use the lowest accuracy as the final uncertainty
        final_uncertainties.append(1 - min(accuracies))

    return final_predictions, final_uncertainties


def confidence_weighted_ensemble(ensemble_preds_ls, labels):
    """Weighted ensemble based on confidence, normalized for all samples per model"""
    final_predictions = []
    final_confidences = []

    for idx in ensemble_preds_ls[0].index:
        class_probabilities = {label: 0 for label in labels}
        total_confidence = 0

        for df in ensemble_preds_ls:
            row = df.loc[idx]
            confidence = 1 - row["uncertainty"]  # confidence score
            total_confidence += confidence

            for label in labels:
                class_probabilities[label] += row[label] * confidence

        # Normalize probabilities
        for label in labels:
            class_probabilities[label] /= total_confidence

        predicted_class = max(class_probabilities, key=class_probabilities.get)
        final_predictions.append(predicted_class)
        final_confidences.append(max(class_probabilities.values()))

    return final_predictions, final_confidences


def performance_weighted_ensemble(ensemble_preds_ls, labels, weights):
    """Weighted ensemble based on performances"""
    final_predictions = []
    final_confidences = []

    # asserts weiths in decimal 0-1
    assert all(0 <= w <= 1 for w in weights), "Weights must be between 0 and 1"

    # assert lenght of weights matches number of models
    assert len(weights) == len(
        ensemble_preds_ls
    ), "Number of weights must match number of models"

    # Normalize weights
    total_weight = sum(weights)
    weights = [weight / total_weight for weight in weights]

    for idx in ensemble_preds_ls[0].index:
        class_probabilities = {label: 0 for label in labels}

        for df, weight in zip(ensemble_preds_ls, weights):
            row = df.loc[idx]
            for label in labels:
                class_probabilities[label] += row[label] * weight

        predicted_class = max(class_probabilities, key=class_probabilities.get)
        final_predictions.append(predicted_class)
        final_confidences.append(max(class_probabilities.values()))

    return final_predictions, final_confidences


def ensemble_predictions(
    config_files: List[str],
    method: str,
    threshold: float,
    weights: List[float],
    no_labels: bool,
) -> pd.DataFrame:
    """
    Ensemble predictions from multiple experiments.

    Args:
        config_files (list): List of configuration file paths.
        method (str): Ensemble method to use. Options are 'majority_voting', 'mean', 'max', or 'sum'.
        no_labels (bool): Flag indicating whether the predictions have labels or not.

    Returns:
        pandas.DataFrame: The ensemble predictions.

    Raises:
        ValueError: If an unknown ensemble method is provided.
        AssertionError: If the number of config files is less than 2 for majority voting.

    """
    ensemble_preds_ls = []
    for config_file in config_files:
        if no_labels:
            # for ensembling results from Nkululeko.demo
            preds = pd.read_csv(config_file)
            labels = preds.columns[1:-2]
        else:
            # for ensembling results from Nkululeko.nkululeko
            config = configparser.ConfigParser()
            config.read(config_file)
            expr = Experiment(config)
            module = "ensemble"
            expr.set_module(module)
            util = Util(module, has_config=True)
            util.debug(
                f"running {expr.name} from config {config_file}, nkululeko version"
                f" {VERSION}"
            )

            # get labels
            labels = expr.util.get_labels()
            # load the experiment
            # get CSV files of predictions
            pred_name = expr.util.get_pred_name()
            util.debug(f"Loading predictions from {pred_name}")
            preds = pd.read_csv(pred_name)

        ensemble_preds_ls.append(preds)

    # pd concate
    ensemble_preds = pd.concat(ensemble_preds_ls, axis=1)

    if method == "majority_voting":
        assert (
            len(ensemble_preds_ls) > 2
        ), "Majority voting only works for more than two models"
        ensemble_preds["predicted"] = majority_voting(ensemble_preds_ls)
    elif method == "mean":
        ensemble_preds["predicted"] = mean_ensemble(ensemble_preds, labels)
    elif method == "max":
        ensemble_preds["predicted"] = max_ensemble(ensemble_preds, labels)
    elif method == "sum":
        ensemble_preds["predicted"] = sum_ensemble(ensemble_preds, labels)
    elif method == "uncertainty":
        ensemble_preds["predicted"] = uncertainty_ensemble(
            ensemble_preds_ls, labels, threshold
        )
    elif method == "uncertainty_weighted":
        (
            ensemble_preds["predicted"],
            ensemble_preds["uncertainty"],
        ) = uncertainty_weighted_ensemble(ensemble_preds_ls, labels)
    elif method == "confidence_weighted":
        (
            ensemble_preds["predicted"],
            ensemble_preds["confidence"],
        ) = confidence_weighted_ensemble(ensemble_preds_ls, labels)
    elif method == "performance_weighted":
        (
            ensemble_preds["predicted"],
            ensemble_preds["confidence"],
        ) = performance_weighted_ensemble(ensemble_preds_ls, labels, weights)
    else:
        raise ValueError(f"Unknown ensemble method: {method}")

    # get the highest value from all labels to infer the label
    # replace the old first predicted column
    if method in ["mean", "max", "sum"]:
        ensemble_preds["predicted"] = ensemble_preds[labels].idxmax(axis=1)

    if no_labels:
        return ensemble_preds

    # Drop start, end columns
    ensemble_preds = ensemble_preds.drop(columns=["start", "end"])

    # Drop other column except until truth
    ensemble_preds = ensemble_preds.iloc[:, : len(labels) + 3]

    # calculate UAR from predicted and truth columns
    truth = ensemble_preds["truth"]
    predicted = ensemble_preds["predicted"]
    uar = balanced_accuracy_score(truth, predicted)
    acc = (truth == predicted).mean()
    # print classification report
    Util("ensemble").debug(f"\n {classification_report(truth, predicted, digits=4)}")
    Util("ensemble").debug(f"{method}: UAR: {uar:.3f}, ACC: {acc:.3f}")

    return ensemble_preds


def main(src_dir: Path) -> None:
    parser = ArgumentParser()
    parser.add_argument(
        "configs",
        nargs="+",
        help="Paths to the configuration files of the experiments to ensemble. \
             Can be INI files for Nkululeko.nkululeko or CSV files from Nkululeko.demo.",
    )
    parser.add_argument(
        "--method",
        default=DEFAULT_METHOD,
        choices=[
            "majority_voting",
            "mean",
            "max",
            "sum",
            # "max_class",
            # "uncertainty_lowest",
            # "entropy",
            "uncertainty",
            "uncertainty_weighted",
            "confidence_weighted",
            "performance_weighted",
        ],
        help=f"Ensemble method to use (default: {DEFAULT_METHOD})",
    )
    # add threshold if method is uncertainty_threshold
    parser.add_argument(
        "--threshold",
        default=1.0,
        type=float,
        help="Threshold for uncertainty_threshold method (default: 1.0, i.e. no threshold)",
    )
    parser.add_argument(
        "--outfile",
        default=DEFAULT_OUTFILE,
        help=f"Output file path for the ensemble predictions (default: {DEFAULT_OUTFILE})",
    )
    parser.add_argument(
        "--weights",
        default=None,
        nargs="+",
        type=float,
        help="Weights for the ensemble method (default: None, e.g. 0.5 0.5)",
    )
    parser.add_argument(
        "--no_labels",
        action="store_true",
        help="True if true labels are not available. For Nkululeko.demo results.",
    )

    args = parser.parse_args()

    start = time.time()

    ensemble_preds = ensemble_predictions(
        args.configs, args.method, args.threshold, args.weights, args.no_labels
    )

    # save to csv
    ensemble_preds.to_csv(args.outfile, index=False)
    Util("ensemble").debug(f"Ensemble predictions saved to: {args.outfile}")
    Util("ensemble").debug(f"Ensemble done, used {time.time()-start:.2f} seconds")

    Util("ensemble").debug("DONE")


if __name__ == "__main__":
    cwd = Path(__file__).parent
    main(cwd)



""""
A predictor for emotional valence.
Currently based on audEERING's emotional dimension model.
"""

import ast

import nkululeko.glob_conf as glob_conf
from nkululeko.feature_extractor import FeatureExtractor
from nkululeko.utils.util import Util


class ValencePredictor:
    """
    ValencePredictor
    predicting valence with the audEERING emotional dimension model

    """

    def __init__(self, df):
        self.df = df
        self.util = Util("valencePredictor")

    def predict(self, split_selection):
        self.util.debug(f"predicting valence for {split_selection} samples")
        feats_name = "_".join(ast.literal_eval(glob_conf.config["DATA"]["databases"]))
        self.feature_extractor = FeatureExtractor(
            self.df, ["auddim"], feats_name, split_selection
        )
        pred_df = self.feature_extractor.extract()
        pred_vals = pred_df.valence * 1000
        return_df = self.df.copy()
        return_df["valence_pred"] = pred_vals.astype("int") / 1000

        return return_df



#!/usr/bin/env python3
# process_database.py --> ENTERFACE


import argparse
from pathlib import Path

import pandas as pd

emotion_map = {
    "an": "anger",
    "di": "disgust",
    "fe": "fear",
    "ha": "happiness",
    "sa": "sadness",
    "su": "surprise",
}


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str, default='ENTERFACE', help='Path to the database file')
    parser.add_argument('--output_dir', type=str, default='.', help='Path to the output directory')
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    if not output_dir.exists():
        output_dir.mkdir()


    paths = list(data_dir.glob('**/*.wav'))

    if len(paths) == 0:
        paths = list(data_dir.glob('**/*.avi'))
    
        # manual correction
        for f in paths:
            if "subject 11" in f.parts:
                f.rename(f.parent / f.name.replace("s12", "s11"))
            if f.stem == 's16_su_3avi':
                f.rename(f.parent / f.name.replace("s16_su_3avi", "s16_su_3.avi"))
            if f.stem.startswith('s_3_'):
                f.rename(f.parent / f.name.replace("s_3", "s3"))

        # reload paths
        paths = list(data_dir.glob('**/*.avi'))
        # pd.DataFrame({"file": paths}).to_csv(output_dir / 'enterface_avi.csv', index=True, index_label='id')

        
    emotions = [p.parts[-2] if p.stem.startswith('s6_') else p.parts[-3] for p in paths]
    speakers = [p.stem.split('_')[0][1:] for p in paths]
    languages = ["english" for p in paths]

    df = pd.DataFrame({'file': paths, 'emotion': emotions, 'speaker': speakers, 'language': languages})


    # allocate speakers > 33 for test
    df_test = df[df['speaker'] > '40']
    df_train = df.drop(df_test.index)


    # save to csv
    df_train.to_csv(output_dir / 'enterface_train_avi.csv', index=False)
    df_test.to_csv(output_dir / 'enterface_test_avi.csv', index=False)
    
    # remove extension, change to wav
    df_train['file'] = df_train['file'].apply(lambda x: x.with_suffix('.wav'))
    df_test.loc[:,'file'] = df_test['file'].apply(lambda x: x.with_suffix('.wav'))

    df_train.to_csv(output_dir / 'enterface_train.csv', index=False)
    df_test.to_csv(output_dir / 'enterface_test.csv', index=False)


    print(f"Total: {len(df)}, Train: {len(df_train)}, Test: {len(df_test)}")


if __name__ == '__main__':
    main()



"""Extract acoustic features from audio samples.

Extract acoustic features using several feature extractors 
(appends the features column-wise)
"""

import pandas as pd

from nkululeko.utils.util import Util


class FeatureExtractor:
    """Extract acoustic features from audio samples.

    Extract acoustic features using several feature extractors (appends the features column-wise).

    Args:
        data_df (pandas.DataFrame): dataframe with audiofile paths as index
        feats_types (List[str]): designations of acoustic feature extractors to be used
        data_name (str): name of databases that are extracted (for caching)
        feats_designation (str): the type of split (train/test), also is used for the cache name.

    Returns:
        df (pandas.DataFrame): dataframe with same index as data_df and acoustic features in columns
    """

    # pandas dataframe to store the features (and indexed with the data from the sets)
    df = None
    data_df = None  # dataframe to get audio paths

    def __init__(self, data_df, feats_types, data_name, feats_designation):
        self.data_df = data_df
        self.data_name = data_name
        self.feats_types = feats_types
        self.util = Util("feature_extractor")
        self.feats_designation = feats_designation

    def extract(self):
        self.feats = pd.DataFrame()
        for feats_type in self.feats_types:
            store_name = f"{self.data_name}_{feats_type}"
            self.feat_extractor = self._get_feat_extractor(store_name, feats_type)
            self.feat_extractor.extract()
            self.feat_extractor.filter()
            self.feats = pd.concat([self.feats, self.feat_extractor.df], axis=1)
        return self.feats

    def extract_sample(self, signal, sr):
        return self.feat_extractor.extract_sample(signal, sr)

    def _get_feat_extractor(self, store_name, feats_type):
        if isinstance(feats_type, list) and len(feats_type) == 1:
            feats_type = feats_type[0]
        feat_extractor_class = self._get_feat_extractor_class(feats_type)
        if feat_extractor_class is None:
            self.util.error(f"unknown feats_type: {feats_type}")
        return feat_extractor_class(
            f"{store_name}_{self.feats_designation}", self.data_df, feats_type
        )

    def _get_feat_extractor_class(self, feats_type):
        if feats_type == "os":
            from nkululeko.feat_extract.feats_opensmile import Opensmileset

            return Opensmileset

        elif feats_type == "spectra":
            from nkululeko.feat_extract.feats_spectra import Spectraloader

            return Spectraloader

        elif feats_type == "trill":
            from nkululeko.feat_extract.feats_trill import TRILLset

            return TRILLset

        elif feats_type.startswith(
            ("wav2vec2", "hubert", "wavlm", "spkrec", "whisper", "ast")
        ):
            return self._get_feat_extractor_by_prefix(feats_type)

        elif feats_type == "xbow":
            from nkululeko.feat_extract.feats_oxbow import Openxbow

            return Openxbow

        elif feats_type in (
            "audmodel",
            "auddim",
            "agender",
            "agender_agender",
            "snr",
            "mos",
            "squim",
            "clap",
            "praat",
            "mld",
            "import",
        ):
            return self._get_feat_extractor_by_name(feats_type)
        else:
            return None

    def _get_feat_extractor_by_prefix(self, feats_type):
        prefix, _, ext = feats_type.partition("-")
        from importlib import import_module

        module = import_module(f"nkululeko.feat_extract.feats_{prefix.lower()}")
        class_name = f"{prefix.capitalize()}"
        return getattr(module, class_name)

    def _get_feat_extractor_by_name(self, feats_type):
        from importlib import import_module

        module = import_module(f"nkululeko.feat_extract.feats_{feats_type.lower()}")
        class_name = f"{feats_type.capitalize()}Set"
        return getattr(module, class_name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# process_database.py --> EMOV-DB

import argparse
import re
from pathlib import Path

import pandas as pd

from nkululeko.utils import files

emotion_map = {
    "amused": "amusement",
    "anger": "anger",
    "disgust": "disgust",
    "neutral": "neutral",
    "sleepiness": "sleepiness",
}

gender_map = {"bea": "female", "jenie": "female", "josh": "male", "sam": "male"}


def main():
    parser = argparse.ArgumentParser(description="Process database")
    parser.add_argument("--data_dir", type=str, default="EmoV-DB_sorted")
    parser.add_argument("--output_dir", type=str, default=".")
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    # paths_list = files.find_files(data_dir, ext="wav")
    paths = list(data_dir.glob("**/*.wav"))
    # paths = [Path(p) for p in paths_list]

    emotions = [emotion_map[p.stem.split("_")[0].lower()] for p in paths]
    speakers = [p.parts[-3] for p in paths]
    genders = [gender_map[s] for s in speakers]
    languages = ["english" for file in paths]

    # convert to df
    df = pd.DataFrame(
        {
            "file": paths,
            "emotion": emotions,
            "speaker": speakers,
            "gender": genders,
            "language": languages,
        }
    )

    # allocate speaker sam for test
    train_df = df[df["speaker"] != "bea"]
    test_df = df.drop(train_df.index)

    # save to csv
    df.to_csv(output_dir / "emov-db.csv", index=False)
    train_df.to_csv(output_dir / "emov-db_train.csv", index=False)
    test_df.to_csv(output_dir / "emov-db_test.csv", index=False)

    print(f"Total: {len(df)}, Train: {len(train_df)}, Test: {len(test_df)}")


if __name__ == "__main__":
    main()



# process_database.py --> ESD

import argparse
import os
from pathlib import Path

import numpy as np
import pandas as pd

emotion_map = {
    'Neutral' : "neutral", 
    'Sad': "sadness",
    'Surprise': "surprise", 
    'Angry': "anger",
    'Happy': "happiness"
}

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", type=str, default="ESD")
    parser.add_argument("--output_dir", type=str, default=".")
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    # Load data
    wav_files = list(data_dir.glob("**/*.wav"))
    
    file = [f for f in wav_files]
    emotion = [f.parent.name for f in wav_files]
    speaker = [f.parts[1] for f in wav_files]
    language = ['chinese' if int(f.parts[1]) <=10 else 'english' for f in wav_files]
    gender = ['female' if int(f.parts[1] in [ [1, 2, 3, 7, 9, 12, 15, 16, 17, 18]])
              else 'male' for f in wav_files]

    # Create dataframe
    df = pd.DataFrame({'file': file, 'emotion': emotion, 'speaker': speaker, 
                       'language': language, 'gender': gender})

    # Split into training and test sets, set speaker 9,10 and 18,20 aside for test
    test_df = df[df['speaker'].isin(['0009', '0010', '0018', '0020'])]
    # drop the rest for test
    train_df = df.drop(test_df.index)

    # save to csv
    train_df.to_csv(output_dir / "esd_train.csv", index=False)
    test_df.to_csv(output_dir / "esd_test.csv", index=False)

if __name__ == "__main__":
    main()


#!/usr/bin/env python3
# process_database.py --> GerPaRas

import argparse
from pathlib import Path

import pandas as pd


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", help="input dir", default="./GerPaRaS/")
    parser.add_argument("--out_dir", help="output output dir", default="./")
    args = parser.parse_args()

    # Read valence, arousal, and dominance from the database
    data_dir = args.data_dir

    val_df = pd.read_csv(data_dir + "df_valence.csv").rename(columns={
        "Media File": "file", "orig": "valence", 'scrambled': 'valence_scrambled'})
    aro_df = pd.read_csv(data_dir + "df_arousal.csv").rename(columns={
        "Media File": "file", "orig": "arousal", 'scrambled': 'arousal_scrambled'})
    dom_df = pd.read_csv(data_dir + "df_dominance.csv").rename(columns={
        "Media File": "file", "orig": "dominance", 'scrambled': 'dominance_scrambled'})

    # standardize original score and scrambled score from [-6, 6] to [0, 1]
    val_df["valence"] = val_df["valence"].apply(lambda x: (x + 6) / 12)
    aro_df["arousal"] = aro_df["arousal"].apply(lambda x: (x + 6) / 12)
    dom_df["dominance"] = dom_df["dominance"].apply(lambda x: (x + 6) / 12)

    val_df["valence_scrambled"] = val_df["valence_scrambled"].apply(lambda x: (x + 6) / 12) 
    aro_df["arousal_scrambled"] = aro_df["arousal_scrambled"].apply(lambda x: (x + 6) / 12)
    dom_df["dominance_scrambled"] = dom_df["dominance_scrambled"].apply(lambda x: (x + 6) / 12)

    # Merge the three dataframes, use `orig` as value for `valence`, `arousal`, and `dominance`
    # so that a file has values of val, aro, and dom for original only
    df = val_df.merge(aro_df, on="file").merge(dom_df, on="file")
    # Drop the `scrambled` columns
    # orig_df = df.drop(columns=["val_scrambled", "aro_scrambled", "dom_scrambled"])

    # get speaker id, this is the first word of the file name before "_"
    df["speaker"] = df["file"].apply(lambda x: x.split("_")[0])

    # rename speaker goering to goering_eckhardt
    df["speaker"] = df["speaker"].apply(lambda x: "goering_eckhardt" if x == "goering" else x)
        
    # add speaker id as parent folder to file name
    df["file"] = df["speaker"] + "/" + df["file"]

    # set speakers gauland and weidel as test set
    test_speakers = ["gauland", "weidel"]

    # set all other speakers as train set
    train_speakers = [s for s in df["speaker"].unique() if s not in test_speakers]

    # set train and test set
    train_df = df[df["speaker"].isin(train_speakers)]
    test_df = df[df["speaker"].isin(test_speakers)]

    # save train and test set to csv
    train_df.to_csv(args.out_dir + "gerparas_train.csv", index=False)
    test_df.to_csv(args.out_dir + "gerparas_test.csv", index=False)
    df.to_csv(args.out_dir + "gerparas.csv", index=False)

    # print length of train and test set and total
    print(f"Total length: {len(df)}, Train set: {len(train_df)}, Test set: {len(test_df)}")


if __name__ == "__main__":
    main()



"""
Resample audio files or INI files (train, test, all) to change the sampling rate.

This script provides a command-line interface to resample audio files or INI files
containing train, test, and all data. It supports resampling a single file, a
directory of files, or all files specified in an INI configuration file.

The script uses the `Resampler` class from the `nkululeko.augmenting.resampler`
module to perform the resampling operation. It can optionally replace the original
audio files with the resampled versions.

The script supports the following command-line arguments:
- `--config`: The base configuration file (INI format) to use for resampling.
- `--file`: The input audio file to resample.
- `--folder`: The input directory containing audio files and subdirectories to resample.
- `--replace`: Whether to replace the original audio files with the resampled versions.

The script also supports loading configuration from an INI file, which can be used
to specify the sample selection (all, train, or test) and whether to replace the
original files.
"""

# resample.py
# change the sampling rate for audio file or INI file (train, test, all)

import argparse
import configparser
import os

import audformat
import pandas as pd

from nkululeko.augmenting.resampler import Resampler
from nkululeko.constants import VERSION
from nkululeko.experiment import Experiment
from nkululeko.utils.files import find_files
from nkululeko.utils.util import Util


def main():
    parser = argparse.ArgumentParser(
        description="Call the nkululeko RESAMPLE framework."
    )
    parser.add_argument("--config", default=None, help="The base configuration")
    parser.add_argument("--file", default=None, help="The input audio file to resample")
    parser.add_argument(
        "--folder",
        default=None,
        help="The input directory containing audio files and subdirectories to resample",
    )
    parser.add_argument(
        "--replace", action="store_true", help="Replace the original audio file"
    )

    args = parser.parse_args()

    if args.file is None and args.folder is None and args.config is None:
        print("ERROR: Either --file, --folder, or --config argument must be provided.")
        exit()

    if args.file is not None:
        # Load the audio file into a DataFrame
        files = pd.Series([args.file])
        df_sample = pd.DataFrame(index=files)
        df_sample.index = audformat.utils.to_segmented_index(
            df_sample.index, allow_nat=False
        )

        # Resample the audio file
        util = Util("resampler", has_config=False)
        util.debug(f"Resampling audio file: {args.file}")
        rs = Resampler(df_sample, not_testing=True, replace=args.replace)
        rs.resample()
    elif args.folder is not None:
        # Load all audio files in the directory and its subdirectories into a DataFrame
        files = find_files(args.folder, relative=True, ext=["wav"])
        files = pd.Series(files)
        df_sample = pd.DataFrame(index=files)
        df_sample.index = audformat.utils.to_segmented_index(
            df_sample.index, allow_nat=False
        )

        # Resample the audio files
        util = Util("resampler", has_config=False)
        util.debug(f"Resampling audio files in directory: {args.folder}")
        rs = Resampler(df_sample, not_testing=True, replace=args.replace)
        rs.resample()
    else:
        # Existing code for handling INI file
        config_file = args.config

        # Test if the configuration file exists
        if not os.path.isfile(config_file):
            print(f"ERROR: no such file: {config_file}")
            exit()

        # Load one configuration per experiment
        config = configparser.ConfigParser()
        config.read(config_file)
        # Create a new experiment
        expr = Experiment(config)
        module = "resample"
        expr.set_module(module)
        util = Util(module)
        util.debug(
            f"running {expr.name} from config {config_file}, nkululeko version"
            f" {VERSION}"
        )

        if util.config_val("EXP", "no_warnings", False):
            import warnings

            warnings.filterwarnings("ignore")

        # Load the data
        expr.load_datasets()

        # Split into train and test
        expr.fill_train_and_tests()
        util.debug(
            f"train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}"
        )

        sample_selection = util.config_val("RESAMPLE", "sample_selection", "all")
        if sample_selection == "all":
            df = pd.concat([expr.df_train, expr.df_test])
        elif sample_selection == "train":
            df = expr.df_train
        elif sample_selection == "test":
            df = expr.df_test
        else:
            util.error(
                f"unknown selection specifier {sample_selection}, should be [all |"
                " train | test]"
            )
        util.debug(f"resampling {sample_selection}: {df.shape[0]} samples")
        replace = util.config_val("RESAMPLE", "replace", "False")
        rs = Resampler(df, replace=replace)
        rs.resample()


if __name__ == "__main__":
    main()



# process_database.py --> VIVAE
# modified from https://github.com/Strong-AI-Lab/emotion/blob/master/datasets/VIVAE/process.py

import argparse
from pathlib import Path

import pandas as pd


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str, default='VIVAE')
    parser.add_argument('--output_dir', type=str, default='.')
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    paths = list(data_dir.glob('full_set/*.wav'))
    emotion = [p.stem.split('_')[1] for p in paths]
    speakers = [p.stem.split('_')[0] for p in paths]
    intensities = [p.stem.split('_')[2] for p in paths]

    df = pd.DataFrame({"file": paths, "emotion": emotion, "speaker": speakers, "intensity": intensities})

    # allocate speaker S05 and S06 to test set
    test_speakers = ['S05', 'S06']
    df_test = df[df['speaker'].isin(test_speakers)]
    df_train = df.drop(df_test.index)

    # save to CSV
    df.to_csv(output_dir / 'vivae.csv', index=False)
    df_train.to_csv(output_dir / 'vivae_train.csv', index=False)
    df_test.to_csv(output_dir / 'vivae_test.csv', index=False)


if __name__ == '__main__':
    main()



# dataset_csv.py
import ast
import os
import os.path

import audformat.utils
import pandas as pd

import nkululeko.glob_conf as glob_conf
from nkululeko.data.dataset import Dataset
from nkululeko.reporting.report_item import ReportItem


class Dataset_CSV(Dataset):
    """Class to represent datasets stored as a csv file"""

    def load(self):
        """Load the dataframe with files, speakers and task labels"""
        self.util.debug(f"loading {self.name}")
        self.got_target, self.got_speaker, self.got_gender = False, False, False
        data_file = self.util.config_val_data(self.name, "", "")
        # if not os.path.isabs(data_file):
        #     exp_root = self.util.config_val("EXP", "root", "")
        #     data_file = os.path.join(exp_root, data_file)
        root = os.path.dirname(data_file)
        audio_path = self.util.config_val_data(self.name, "audio_path", "./")
        df = pd.read_csv(data_file)
        # trim all string values
        df_obj = df.select_dtypes("object")
        df[df_obj.columns] = df_obj.apply(lambda x: x.str.strip())
        # special treatment for segmented dataframes with only one column:
        if "start" in df.columns and len(df.columns) == 4:
            index = audformat.segmented_index(
                df.file.values, df.start.values, df.end.values
            )
            df = df.set_index(index)
            df = df.drop(columns=["file", "start", "end"])
        else:
            df = audformat.utils.read_csv(data_file)
        if isinstance(df, pd.Series):
            df = df.to_frame()
        rename_cols = self.util.config_val_data(self.name, "colnames", False)
        if rename_cols:
            col_dict = ast.literal_eval(rename_cols)
            df = df.rename(columns=col_dict)
        absolute_path = eval(
            self.util.config_val_data(self.name, "absolute_path", "True")
        )
        if not absolute_path:
            # add the root folder to the relative paths of the files
            if audformat.index_type(df.index) == "segmented":
                file_index = (
                    df.index.levels[0]
                    .map(lambda x: root + "/" + audio_path + "/" + x)
                    .values
                )
                df = df.set_index(df.index.set_levels(file_index, level="file"))
            else:
                if not isinstance(df, pd.DataFrame):
                    df = pd.DataFrame(df)
                df = df.set_index(
                    df.index.to_series().apply(
                        lambda x: root + "/" + audio_path + "/" + x
                    )
                )
        else:  # absolute path is True
            if audformat.index_type(df.index) == "segmented":
                file_index = (
                    df.index.levels[0].map(lambda x: audio_path + "/" + x).values
                )
                df = df.set_index(df.index.set_levels(file_index, level="file"))
            else:
                if not isinstance(df, pd.DataFrame):
                    df = pd.DataFrame(df)
                df = df.set_index(
                    df.index.to_series().apply(lambda x: audio_path + "/" + x)
                )

        self.df = df
        self.db = None
        self.got_target = True
        self.is_labeled = self.got_target
        self.start_fresh = eval(self.util.config_val("DATA", "no_reuse", "False"))
        is_index = False
        try:
            if self.is_labeled and "class_label" not in self.df.columns:
                self.df["class_label"] = self.df[self.target]
        except AttributeError:
            is_index = True
            r_string = (
                f"Loaded database {self.name} with {df.shape[0]} samples as index."
            )
        if not is_index:
            if "gender" in self.df.columns:
                self.got_gender = True
            if "age" in self.df.columns:
                self.got_age = True
            if "speaker" in self.df.columns:
                self.got_speaker = True
            speaker_num = 0
            if self.got_speaker:
                speaker_num = self.df.speaker.nunique()
            r_string = (
                f"Loaded database {self.name} with {df.shape[0]} samples: got"
                f" targets: {self.got_target}, got speakers:"
                f" {self.got_speaker} ({speaker_num}), got sexes:"
                f" {self.got_gender}, got age: {self.got_age}"
            )
        self.util.debug(r_string)
        glob_conf.report.add_item(ReportItem("Data", "Loaded report", r_string))

    def prepare(self):
        super().prepare()



"""Code based on @jwagner."""

import ast
import dataclasses
import json
import os
import pickle
import typing

import audeer
import audiofile
import audmetric
import datasets
import numpy as np
import pandas as pd
import torch
import transformers
from transformers.models.wav2vec2.modeling_wav2vec2 import (
    Wav2Vec2Model,
    Wav2Vec2PreTrainedModel,
)

import nkululeko.glob_conf as glob_conf
from nkululeko.models.model import Model as BaseModel
from nkululeko.reporting.reporter import Reporter


class TunedModel(BaseModel):
    def __init__(self, df_train, df_test, feats_train, feats_test):
        """Constructor taking the configuration and all dataframes."""
        super().__init__(df_train, df_test, feats_train, feats_test)
        super().set_model_type("finetuned")
        self.df_test, self.df_train, self.feats_test, self.feats_train = (
            df_test,
            df_train,
            feats_test,
            feats_train,
        )
        self.name = "finetuned_wav2vec2"
        self.target = glob_conf.config["DATA"]["target"]
        self.labels = glob_conf.labels
        self.class_num = len(self.labels)
        device = self.util.config_val("MODEL", "device", False)
        if not device:
            self.device = "cuda" if torch.cuda.is_available() else "cpu"
        else:
            self.device = device
        if self.device != "cpu":
            os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
            os.environ["CUDA_VISIBLE_DEVICES"] = self.device
        self.util.debug(f"running on device {self.device}")
        self.is_classifier = self.util.exp_is_classification()
        if self.is_classifier:
            self.measure = "uar"
        else:
            self.measure = self.util.config_val("MODEL", "measure", "ccc")
        self.util.debug(f"evaluation metrics: {self.measure}")
        self.batch_size = int(self.util.config_val("MODEL", "batch_size", "8"))
        self.util.debug(f"batch size: {self.batch_size}")
        self.learning_rate = float(
            self.util.config_val("MODEL", "learning_rate", "0.0001")
        )
        self.max_duration = float(self.util.config_val("MODEL", "max_duration", "8.0"))
        self.df_train, self.df_test = df_train, df_test
        self.epoch_num = int(self.util.config_val("EXP", "epochs", 1))
        self.util.debug(f"num of epochs: {self.epoch_num}")
        drop = self.util.config_val("MODEL", "drop", False)
        self.drop = 0.1
        if drop:
            self.drop = float(drop)
        self.util.debug(f"init: training with dropout: {self.drop}")
        self.push = eval(self.util.config_val("MODEL", "push_to_hub", "False"))
        self.balancing = self.util.config_val("MODEL", "balancing", False)
        self._init_model()

    def _init_model(self):
        model_path = "facebook/wav2vec2-large-robust-ft-swbd-300h"
        pretrained_model = self.util.config_val("MODEL", "pretrained_model", model_path)
        self.num_layers = None
        self.sampling_rate = 16000
        self.max_duration_sec = self.max_duration
        self.accumulation_steps = 4

        # print finetuning information via debug
        self.util.debug(f"Finetuning from model: {pretrained_model}")

        # create dataset
        dataset = {}
        target_name = glob_conf.target
        data_sources = {
            "train": pd.DataFrame(self.df_train[target_name]),
            "dev": pd.DataFrame(self.df_test[target_name]),
        }

        for split in ["train", "dev"]:
            df = data_sources[split]
            y = df[target_name].astype("float")
            y.name = "targets"
            df = y.reset_index()
            df.start = df.start.dt.total_seconds()
            df.end = df.end.dt.total_seconds()
            #     ds = datasets.Dataset.from_pandas(df)
            #     dataset[split] = ds

            # self.dataset = datasets.DatasetDict(dataset)
            if split == "train" and self.balancing:
                if self.balancing == "ros":
                    from imblearn.over_sampling import RandomOverSampler

                    sampler = RandomOverSampler(random_state=42)
                elif self.balancing == "smote":
                    from imblearn.over_sampling import SMOTE

                    sampler = SMOTE(random_state=42)
                elif self.balancing == "adasyn":
                    from imblearn.over_sampling import ADASYN

                    sampler = ADASYN(random_state=42)
                else:
                    self.util.error(f"Unknown balancing algorithm: {self.balancing}")

                X_resampled, y_resampled = sampler.fit_resample(
                    df[["start", "end"]], df["targets"]
                )
                df = pd.DataFrame(
                    {
                        "start": X_resampled["start"],
                        "end": X_resampled["end"],
                        "targets": y_resampled,
                    }
                )

                # print the before and after class distribution
                self.util.debug(
                    f"balanced with: {self.balancing}, new size: {len(df)}, was {len(data_sources[split])}"
                )

            ds = datasets.Dataset.from_pandas(df)
            dataset[split] = ds

        self.dataset = datasets.DatasetDict(dataset)

        # load pre-trained model
        if self.is_classifier:
            self.util.debug("Task is classification.")
            le = glob_conf.label_encoder
            mapping = dict(zip(le.classes_, range(len(le.classes_))))
            target_mapping = {k: int(v) for k, v in mapping.items()}
            target_mapping_reverse = {
                value: key for key, value in target_mapping.items()
            }
            self.config = transformers.AutoConfig.from_pretrained(
                pretrained_model,
                num_labels=len(target_mapping),
                label2id=target_mapping,
                id2label=target_mapping_reverse,
                finetuning_task=target_name,
            )
        else:
            self.util.debug("Task is regression.")
            self.config = transformers.AutoConfig.from_pretrained(
                pretrained_model,
                num_labels=1,
                finetuning_task=target_name,
            )
        if self.num_layers is not None:
            self.config.num_hidden_layers = self.num_layers
        self.config.final_dropout = self.drop
        setattr(self.config, "sampling_rate", self.sampling_rate)
        setattr(self.config, "data", self.util.get_data_name())
        setattr(self.config, "is_classifier", self.is_classifier)

        vocab_dict = {}
        with open("vocab.json", "w") as vocab_file:
            json.dump(vocab_dict, vocab_file)
        tokenizer = transformers.Wav2Vec2CTCTokenizer("./vocab.json")
        tokenizer.save_pretrained(".")

        # uoload tokenizer to hub if true
        if self.push:
            tokenizer.push_to_hub(self.util.get_name())

        feature_extractor = transformers.Wav2Vec2FeatureExtractor(
            feature_size=1,
            sampling_rate=16000,
            padding_value=0.0,
            do_normalize=True,
            return_attention_mask=True,
        )
        self.processor = transformers.Wav2Vec2Processor(
            feature_extractor=feature_extractor,
            tokenizer=tokenizer,
        )
        assert self.processor.feature_extractor.sampling_rate == self.sampling_rate

        self.model = Model.from_pretrained(
            pretrained_model,
            config=self.config,
        )
        self.model.freeze_feature_extractor()
        self.model.train()
        self.model_initialized = True

    def set_model_type(self, type):
        self.model_type = type

    def set_testdata(self, data_df, feats_df):
        self.df_test, self.feats_test = data_df, feats_df

    def reset_test(self, df_test, feats_test):
        self.df_test, self.feats_test = df_test, feats_test

    def set_id(self, run, epoch):
        self.run = run
        self.epoch = epoch
        dir = self.util.get_path("model_dir")
        name = f"{self.util.get_exp_name(only_train=True)}_{self.run}_{self.epoch:03d}.model"
        self.store_path = dir + name

    def data_collator(self, data):
        files = [d["file"] for d in data]
        starts = [d["start"] for d in data]
        ends = [d["end"] for d in data]
        targets = [d["targets"] for d in data]

        signals = []
        for file, start, end in zip(
            files,
            starts,
            ends,
        ):
            offset = start
            duration = end - offset
            if self.max_duration_sec is not None:
                duration = min(duration, self.max_duration_sec)
            signal, _ = audiofile.read(
                file,
                offset=offset,
                duration=duration,
            )
            signals.append(signal.squeeze())

        input_values = self.processor(
            signals,
            sampling_rate=self.sampling_rate,
            padding=True,
        )
        batch = self.processor.pad(
            input_values,
            padding=True,
            return_tensors="pt",
        )

        batch["labels"] = torch.Tensor(targets)

        return batch

    def compute_metrics(self, p: transformers.EvalPrediction):
        metrics = {
            "UAR": audmetric.unweighted_average_recall,
            "ACC": audmetric.accuracy,
        }
        metrics_reg = {
            "PCC": audmetric.pearson_cc,
            "CCC": audmetric.concordance_cc,
            "MSE": audmetric.mean_squared_error,
            "MAE": audmetric.mean_absolute_error,
        }

        # truth = p.label_ids[:, 0].astype(int)
        truth = p.label_ids
        preds = p.predictions
        preds = np.argmax(preds, axis=1)
        scores = {}
        if self.is_classifier:
            for name, metric in metrics.items():
                scores[f"{name}"] = metric(truth, preds)
        else:
            for name, metric in metrics_reg.items():
                scores[f"{name}"] = metric(truth, preds)

        return scores

    def train(self):
        """Train the model."""
        model_root = self.util.get_path("model_dir")
        self.log_root = os.path.join(self.util.get_exp_dir(), "log")
        audeer.mkdir(self.log_root)
        self.torch_root = audeer.path(model_root, "torch")
        conf_file = os.path.join(self.torch_root, "config.json")
        if os.path.isfile(conf_file):
            self.util.debug(f"reusing finetuned model: {conf_file}")
            self.load(self.run, self.epoch_num)
            return
        targets = pd.DataFrame(self.dataset["train"]["targets"])

        if self.is_classifier:
            criterion = self.util.config_val("MODEL", "loss", "cross")
            if criterion == "cross":
                if self.util.config_val("MODEL", "class_weight", False):
                    counts = targets[0].value_counts().sort_index()
                    train_weights = 1 / counts
                    train_weights /= train_weights.sum()
                    self.util.debug(f"train weights: {train_weights}")
                    criterion = torch.nn.CrossEntropyLoss(
                        weight=torch.Tensor(train_weights).to("cuda"),
                    )
                else:
                    criterion = torch.nn.CrossEntropyLoss()
            else:
                self.util.error(f"criterion {criterion} not supported for classifier")
        else:
            criterion = self.util.config_val("MODEL", "loss", "1-ccc")
            if criterion == "1-ccc":
                criterion = ConcordanceCorCoeff()
            elif criterion == "mse":
                criterion = torch.nn.MSELoss()
            elif criterion == "mae":
                criterion = torch.nn.L1Loss()
            else:
                self.util.error(f"criterion {criterion} not supported for regressor")

        # set push_to_hub value, default false
        # push = eval(self.util.config_val("MODEL", "push_to_hub", "False"))

        class Trainer(transformers.Trainer):
            def compute_loss(
                self,
                model,
                inputs,
                return_outputs=False,
            ):
                targets = inputs.pop("labels").squeeze()
                targets = targets.type(torch.long)

                outputs = model(**inputs)
                logits = outputs[0].squeeze()

                loss = criterion(logits, targets)

                return (loss, outputs) if return_outputs else loss

        num_steps = (
            len(self.dataset["train"])
            // (self.batch_size * self.accumulation_steps)
            // 5
        )
        num_steps = max(1, num_steps)

        metrics_for_best_model = self.measure.upper()
        if metrics_for_best_model == "UAR":
            greater_is_better = True
        elif metrics_for_best_model == "CCC":
            greater_is_better = True
        elif metrics_for_best_model == "MSE":
            greater_is_better = False
        elif metrics_for_best_model == "MAE":
            greater_is_better = False
        else:
            self.util.error(f"unknown metric/measure: {metrics_for_best_model}")

        training_args = transformers.TrainingArguments(
            output_dir=model_root,
            logging_dir=self.log_root,
            per_device_train_batch_size=self.batch_size,
            per_device_eval_batch_size=self.batch_size,
            gradient_accumulation_steps=self.accumulation_steps,
            evaluation_strategy="steps",
            num_train_epochs=self.epoch_num,
            fp16=self.device != "cpu",
            use_cpu=self.device == "cpu",
            save_steps=num_steps,
            eval_steps=num_steps,
            logging_steps=num_steps,
            logging_strategy="epoch",
            learning_rate=self.learning_rate,
            save_total_limit=2,
            metric_for_best_model=metrics_for_best_model,
            greater_is_better=greater_is_better,
            load_best_model_at_end=True,
            remove_unused_columns=False,
            report_to="none",
            push_to_hub=self.push,
            hub_model_id=f"{self.util.get_name()}",
            overwrite_output_dir=True,
        )

        trainer = Trainer(
            model=self.model,
            data_collator=self.data_collator,
            args=training_args,
            compute_metrics=self.compute_metrics,
            train_dataset=self.dataset["train"],
            eval_dataset=self.dataset["dev"],
            tokenizer=self.processor.feature_extractor,
            callbacks=[transformers.integrations.TensorBoardCallback()],
        )

        trainer.train()
        trainer.save_model(self.torch_root)
        log_file = os.path.join(
            self.log_root,
            "log.txt",
        )
        with open(log_file, "w") as text_file:
            print(trainer.state.log_history, file=text_file)
        self.util.debug(f"saved best model to {self.torch_root}")
        self.load(self.run, self.epoch)

    def get_predictions(self):
        results = [[]].pop(0)
        for (file, start, end), _ in audeer.progress_bar(
            self.df_test.iterrows(),
            total=len(self.df_test),
            desc=f"Predicting {len(self.df_test)} audiofiles",
        ):
            if end == pd.NaT:
                signal, sr = audiofile.read(file, offset=start)
            else:
                signal, sr = audiofile.read(
                    file, duration=end - start, offset=start, always_2d=True
                )
            assert sr == self.sampling_rate
            prediction = self.model.predict(signal)
            results.append(prediction)
            # results.append(predictions.argmax())
        predictions = np.asarray(results)
        if self.util.exp_is_classification():
            # make a dataframe for the class probabilities
            proba_d = {}
            for c in range(self.class_num):
                proba_d[c] = []
            # get the class probabilities
            # predictions = self.clf.predict_proba(self.feats_test.to_numpy())
            # pred = self.clf.predict(features)
            for i in range(self.class_num):
                proba_d[i] = list(predictions.T[i])
            probas = pd.DataFrame(proba_d)
            probas = probas.set_index(self.df_test.index)
            predictions = probas.idxmax(axis=1).values
        else:
            predictions = predictions.flatten()
            probas = None
        return predictions, probas

    def predict(self):
        """Predict the whole eval feature set"""
        predictions, probas = self.get_predictions()
        report = Reporter(
            self.df_test[self.target].to_numpy().astype(float),
            predictions,
            self.run,
            self.epoch_num,
            probas=probas,
        )
        self._plot_epoch_progression(report)
        return report

    def _plot_epoch_progression(self, report):
        log_file = os.path.join(
            self.log_root,
            "log.txt",
        )
        with open(log_file, "r") as file:
            data = file.read()
        data = data.strip().replace("nan", "0")
        list = ast.literal_eval(data)
        epochs, vals, loss = [], [], []
        for index, tp in enumerate(list):
            try:
                epochs.append(tp["epoch"])
                measure = self.measure.upper()
                vals.append(tp[f"eval_{measure}"])
                loss.append(tp["eval_loss"])
            except KeyError:
                del epochs[-1]
                # print(f'no value at {index}')
        df = pd.DataFrame({"results": vals, "losses": loss}, index=epochs)
        report.plot_epoch_progression_finetuned(df)

    def predict_sample(self, signal):
        """Predict one sample"""
        prediction = {}
        if self.is_classifier:
            # get the class probabilities
            predictions = self.model.predict(signal)
            # pred = self.clf.predict(features)
            for i in range(len(self.labels)):
                cat = self.labels[i]
                prediction[cat] = predictions[i]
        else:
            predictions = self.model.predict(signal)
            prediction = predictions
        return prediction

    def store(self):
        self.util.debug("stored: ")

    def load(self, run, epoch):
        self.set_id(run, epoch)
        self.model = Model.from_pretrained(
            self.torch_root,
            config=self.config,
        )
        # print(f"loaded model type {type(self.model)}")

    def load_path(self, path, run, epoch):
        self.set_id(run, epoch)
        with open(path, "rb") as handle:
            self.clf = pickle.load(handle)


@dataclasses.dataclass
class ModelOutput(transformers.file_utils.ModelOutput):
    logits: torch.FloatTensor = None
    hidden_states: typing.Tuple[torch.FloatTensor] = None
    cnn_features: torch.FloatTensor = None


@dataclasses.dataclass
class ModelOutputReg(transformers.file_utils.ModelOutput):
    logits: torch.FloatTensor
    hidden_states: typing.Tuple[torch.FloatTensor] = None
    attentions: typing.Tuple[torch.FloatTensor] = None
    logits_framewise: torch.FloatTensor = None
    hidden_states_framewise: torch.FloatTensor = None
    cnn_features: torch.FloatTensor = None


class ModelHead(torch.nn.Module):
    def __init__(self, config):
        super().__init__()

        self.dense = torch.nn.Linear(config.hidden_size, config.hidden_size)
        self.dropout = torch.nn.Dropout(config.final_dropout)
        self.out_proj = torch.nn.Linear(config.hidden_size, config.num_labels)

    def forward(self, features, **kwargs):
        x = features
        x = self.dropout(x)
        x = self.dense(x)
        x = torch.tanh(x)
        x = self.dropout(x)
        x = self.out_proj(x)

        return x


class Model(Wav2Vec2PreTrainedModel):
    def __init__(self, config):
        if not hasattr(config, "add_adapter"):
            setattr(config, "add_adapter", False)

        super().__init__(config)

        self.wav2vec2 = Wav2Vec2Model(config)
        self.head = ModelHead(config)
        self.is_classifier = config.is_classifier
        self.init_weights()

    def freeze_feature_extractor(self):
        self.wav2vec2.feature_extractor._freeze_parameters()

    def pooling(
        self,
        hidden_states,
        attention_mask,
    ):
        if attention_mask is None:  # For evaluation with batch_size==1
            outputs = torch.mean(hidden_states, dim=1)
        else:
            attention_mask = self._get_feature_vector_attention_mask(
                hidden_states.shape[1],
                attention_mask,
            )
            hidden_states = hidden_states * torch.reshape(
                attention_mask,
                (-1, attention_mask.shape[-1], 1),
            )
            outputs = torch.sum(hidden_states, dim=1)
            attention_sum = torch.sum(attention_mask, dim=1)

            epsilon = 1e-6  # to avoid division by zero and numerical instability
            outputs = outputs / (torch.reshape(attention_sum, (-1, 1)) + epsilon)

        return outputs

    def forward(
        self,
        input_values,
        attention_mask=None,
        labels=None,
        return_hidden=False,
    ):
        outputs = self.wav2vec2(
            input_values,
            attention_mask=attention_mask,
        )
        cnn_features = outputs.extract_features
        hidden_states_framewise = outputs.last_hidden_state
        hidden_states = self.pooling(
            hidden_states_framewise,
            attention_mask,
        )
        logits = self.head(hidden_states)
        if not self.training:
            logits = torch.softmax(logits, dim=1)

        if return_hidden:
            # make time last axis
            cnn_features = torch.transpose(cnn_features, 1, 2)
            if self.is_classifier:
                return ModelOutput(
                    logits=logits,
                    hidden_states=hidden_states,
                    cnn_features=cnn_features,
                )
            else:
                return ModelOutputReg(
                    logits=logits,
                    hidden_states=hidden_states,
                    cnn_features=cnn_features,
                )
        else:
            if self.is_classifier:
                return ModelOutput(
                    logits=logits,
                )
            else:
                return ModelOutputReg(
                    logits=logits,
                )

    def predict(self, signal):
        result = self(torch.from_numpy(signal))
        result = result[0].detach().numpy()[0]
        return result


class ConcordanceCorCoeff(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.mean = torch.mean
        self.var = torch.var
        self.sum = torch.sum
        self.sqrt = torch.sqrt
        self.std = torch.std

    def forward(self, prediction, ground_truth):
        ground_truth = ground_truth.float()
        mean_gt = self.mean(ground_truth, 0)
        mean_pred = self.mean(prediction, 0)
        var_gt = self.var(ground_truth, 0)
        var_pred = self.var(prediction, 0)
        v_pred = prediction - mean_pred
        v_gt = ground_truth - mean_gt
        cor = self.sum(v_pred * v_gt) / (
            self.sqrt(self.sum(v_pred**2)) * self.sqrt(self.sum(v_gt**2))
        )
        sd_gt = self.std(ground_truth)
        sd_pred = self.std(prediction)
        numerator = 2 * cor * sd_gt * sd_pred
        denominator = var_gt + var_pred + (mean_gt - mean_pred) ** 2
        ccc = numerator / denominator

        return 1 - ccc



"""
Demonstrates the usage of the ML-experiment framework for the nkululeko MULTIDB project.

The `main` function is the entry point of the script, which parses command-line arguments, reads a configuration file, and runs the nkululeko or aug_train functions based on the configuration.

The `plot_heatmap` function generates a heatmap plot of the results and saves it to a file, along with some summary statistics.
"""

# main.py
# Demonstration code to use the ML-experiment framework

import argparse
import ast
import configparser
import os

import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sn

from nkululeko.aug_train import doit as aug_train
from nkululeko.nkululeko import doit as nkulu


def main(src_dir):
    parser = argparse.ArgumentParser(
        description="Call the nkululeko MULTIDB framework."
    )
    parser.add_argument("--config", default="exp.ini", help="The base configuration")
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config
    else:
        config_file = f"{src_dir}/exp.ini"

    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f"ERROR: no such file: {config_file}")
        exit()

    config = configparser.ConfigParser()
    config.read(config_file)
    datasets = config["EXP"]["databases"]
    datasets = ast.literal_eval(datasets)
    try:
        use_splits = eval(config["EXP"]["use_splits"])
    except KeyError:
        use_splits = False
    dim = len(datasets)
    results = np.zeros(dim * dim).reshape([dim, dim])
    last_epochs = np.zeros(dim * dim).reshape([dim, dim])
    # check if some data should be added to training
    try:
        extra_trains = config["CROSSDB"]["train_extra"]
    except KeyError:
        extra_trains = False

    for i in range(dim):
        for j in range(dim):
            # initialize config
            config = None
            config = configparser.ConfigParser()
            config.read(config_file)
            if i == j:
                dataset = datasets[i]
                print(f"running {dataset}")
                if extra_trains:
                    extra_trains_1 = extra_trains.removeprefix("[").removesuffix("]")
                    config["DATA"]["databases"] = f"['{dataset}', {extra_trains_1}]"
                    extra_trains_2 = ast.literal_eval(extra_trains)
                    for extra_train in extra_trains_2:
                        config["DATA"][f"{extra_train}.split_strategy"] = "train"
                else:
                    config["DATA"]["databases"] = f"['{dataset}']"
                config["EXP"]["name"] = dataset
            else:
                train = datasets[i]
                test = datasets[j]
                print(f"running train: {train}, test: {test}")
                if extra_trains:
                    extra_trains_1 = extra_trains.removeprefix("[").removesuffix("]")
                    config["DATA"][
                        "databases"
                    ] = f"['{train}', '{test}', {extra_trains_1}]"
                    if use_splits:
                        config["DATA"][f"{test}.as_test"] = "True"
                        config["DATA"][f"{train}.as_train"] = "True"
                    else:
                        config["DATA"][f"{test}.split_strategy"] = "test"
                        config["DATA"][f"{train}.split_strategy"] = "train"
                    extra_trains_2 = ast.literal_eval(extra_trains)
                    for extra_train in extra_trains_2:
                        config["DATA"][f"{extra_train}.split_strategy"] = "train"
                else:
                    config["DATA"]["databases"] = f"['{train}', '{test}']"
                    if use_splits:
                        config["DATA"][f"{test}.as_test"] = "True"
                        config["DATA"][f"{train}.as_train"] = "True"
                    else:
                        config["DATA"][f"{test}.split_strategy"] = "test"
                        config["DATA"][f"{train}.split_strategy"] = "train"
                config["EXP"]["name"] = f"{train}_vs_{test}"

            tmp_config = "tmp.ini"
            with open(tmp_config, "w") as tmp_file:
                config.write(tmp_file)
            if config.has_section("AUGMENT"):
                result, last_epoch = aug_train(tmp_config)
            else:
                result, last_epoch = nkulu(tmp_config)
            results[i, j] = float(result)
            last_epochs[i, j] = int(last_epoch)
    print(repr(results))
    print(repr(last_epochs))
    root = os.path.join(config["EXP"]["root"], "")
    try:
        format = config["PLOT"]["format"]
        plot_name = f"{root}/heatmap.{format}"
    except KeyError:
        plot_name = f"{root}/heatmap.png"
    plot_heatmap(results, last_epochs, datasets, plot_name, config, datasets)


def trunc_to_three(x):
    return int(x * 1000) / 1000.0


def plot_heatmap(results, last_epochs, labels, name, config, datasets):
    df_cm = pd.DataFrame(
        results, index=[i for i in labels], columns=[i for i in labels]
    )
    mean = trunc_to_three(results.mean())
    mean_diag = trunc_to_three(results.diagonal().mean())
    mean_non_diag = trunc_to_three(
        (results.sum() - results.diagonal().sum())
        / (results.size - results.diagonal().size)
    )
    colsums = results.mean(axis=0)
    vfunc = np.vectorize(trunc_to_three)
    colsums = vfunc(colsums)
    rowsums = results.mean(axis=1)
    rowsums = vfunc(rowsums)
    colsums_epochs = last_epochs.mean(axis=0)
    colsums_epochs = vfunc(colsums_epochs)
    res_dir = config["EXP"]["root"]
    file_name = f"{res_dir}/results.txt"
    with open(file_name, "w") as text_file:
        text_file.write(
            f"Mean UAR: {mean} (self: {mean_diag}, cross: {mean_non_diag})\n"
        )
        data_s = ", ".join(datasets)
        text_file.write(f"{data_s}\n")
        colsums = np.array2string(colsums, separator=", ")
        text_file.write(f"column means\n{colsums}\n")
        rowsums = np.array2string(rowsums, separator=", ")
        text_file.write(f"rows means\n{rowsums}\n")
        text_file.write("all results\n")
        text_file.write(repr(results))
        text_file.write("\n")
        colsums_epochs = np.array2string(colsums_epochs, separator=", ")
        text_file.write(f"column sums epochs\n{colsums_epochs}\n")
        text_file.write("all epochs\n")
        text_file.write(repr(last_epochs))

    plt.figure(figsize=(10, 7))
    ax = sn.heatmap(df_cm, annot=True, cmap=cm.Blues)
    caption = f"Rows: train, Cols: test. Mean UAR: {mean} (self: {mean_diag}, cross: {mean_non_diag})."
    ax.set_title(caption)
    plt.savefig(name)
    plt.close()


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd)  # sys.argv[1])



import ast
import glob
import json
import math

import matplotlib.pyplot as plt
import numpy as np

# from torch import is_tensor
from audmetric import (
    accuracy,
    concordance_cc,
    mean_absolute_error,
    mean_squared_error,
    unweighted_average_recall,
)

# import os
from confidence_intervals import evaluate_with_conf_int
from scipy.special import softmax
from scipy.stats import entropy, pearsonr
from sklearn.metrics import (
    ConfusionMatrixDisplay,
    RocCurveDisplay,
    auc,
    classification_report,
    confusion_matrix,
    r2_score,
    roc_auc_score,
    roc_curve,
)

import nkululeko.glob_conf as glob_conf
from nkululeko.plots import Plots
from nkululeko.reporting.defines import Header
from nkululeko.reporting.report_item import ReportItem
from nkululeko.reporting.result import Result
from nkululeko.utils.util import Util


class Reporter:
    def _set_metric(self):
        if self.util.exp_is_classification():
            self.metric = "uar"
            self.METRIC = "UAR"
            self.result.metric = self.METRIC
            self.is_classification = True
        else:
            self.is_classification = False
            self.metric = self.util.config_val("MODEL", "measure", "mse")
            if self.metric == "mse":
                self.METRIC = "MSE"
                self.result.metric = self.METRIC
            elif self.metric == "mae":
                self.METRIC = "MAE"
                self.result.metric = self.METRIC
            elif self.metric == "ccc":
                self.METRIC = "CCC"
                self.result.metric = self.METRIC

    def __init__(self, truths, preds, run, epoch, probas=None):
        """Initialization with ground truth und predictions vector.

        Args:
            truths (list): the ground truth
            preds (list): the predictions
            run (int): number of run
            epoch (int): number of epoch
            probas (pd.Dataframe, optional): probabilities per class. Defaults to None.
        """
        self.util = Util("reporter")
        self.probas = probas
        self.format = self.util.config_val("PLOT", "format", "png")
        self.truths = np.asarray(truths)
        self.preds = np.asarray(preds)
        self.result = Result(0, 0, 0, 0, "unknown")
        self.run = run
        self.epoch = epoch
        self.model_type = self.util.get_model_type()
        self._set_metric()
        self.filenameadd = ""
        self.cont_to_cat = False
        if len(self.truths) > 0 and len(self.preds) > 0:
            if self.util.exp_is_classification():
                uar, upper, lower = self._get_test_result(
                    self.truths, self.preds, "uar"
                )
                self.result.test = uar
                self.result.set_upper_lower(upper, lower)
                self.result.loss = 1 - accuracy(self.truths, self.preds)
            else:
                # regression experiment
                # keep the original values for further use, they will be binned later
                self.truths_cont = self.truths
                self.preds_cont = self.preds
                test_result, upper, lower = self._get_test_result(
                    self.truths, self.preds, self.metric
                )
                self.result.test = test_result
                self.result.set_upper_lower(upper, lower)
                # train and loss are being set by the model

    def _get_test_result(self, truths, preds, metric):
        if metric == "uar":
            test_result, (upper, lower) = evaluate_with_conf_int(
                preds,
                unweighted_average_recall,
                truths,
                num_bootstraps=1000,
                alpha=5,
            )
        elif metric == "mse":
            test_result, (upper, lower) = evaluate_with_conf_int(
                preds,
                mean_squared_error,
                truths,
                num_bootstraps=1000,
                alpha=5,
            )
        elif metric == "mae":
            test_result, (upper, lower) = evaluate_with_conf_int(
                preds,
                mean_absolute_error,
                truths,
                num_bootstraps=1000,
                alpha=5,
            )
        elif metric == "ccc":
            test_result, (upper, lower) = evaluate_with_conf_int(
                preds,
                concordance_cc,
                truths,
                num_bootstraps=1000,
                alpha=5,
            )
            if math.isnan(test_result):
                self.util.debug(f"Truth: {self.truths}")
                self.util.debug(f"Predict.: {self.preds}")
                self.util.debug("Result is NAN: setting to -1")
                test_result = -1
        else:
            self.util.error(f"unknown metric: {self.metric}")
        return test_result, upper, lower

    def print_probabilities(self):
        """Print the probabilities per class to a file in the store."""
        if (
            self.util.exp_is_classification()
            and self.probas is not None
            and "uncertainty" not in self.probas
        ):
            probas = self.probas
            # softmax the probabilities or logits
            uncertainty = probas.apply(softmax, axis=1)
            probas["predicted"] = self.preds
            probas["truth"] = self.truths
            try:
                le = glob_conf.label_encoder
                mapping = dict(zip(le.classes_, range(len(le.classes_))))
                mapping_reverse = {value: key for key, value in mapping.items()}
                probas = probas.rename(columns=mapping_reverse)
                probas["predicted"] = probas["predicted"].map(mapping_reverse)
                probas["truth"] = probas["truth"].map(mapping_reverse)
            except AttributeError as ae:
                self.util.debug(f"Can't label categories: {ae}")
            # compute entropy per sample
            uncertainty = uncertainty.apply(entropy)
            # scale it to 0-1
            max_ent = math.log(len(glob_conf.labels))
            uncertainty = (uncertainty - uncertainty.min()) / (
                max_ent - uncertainty.min()
            )
            probas["uncertainty"] = uncertainty
            probas["correct"] = probas.predicted == probas.truth
            sp = self.util.get_pred_name()

            self.probas = probas
            probas.to_csv(sp)
            self.util.debug(f"Saved probabilities to {sp}")
            plots = Plots()
            ax, caption = plots.plotcatcont(
                probas, "correct", "uncertainty", "uncertainty", "correct"
            )
            plots.save_plot(
                ax,
                caption,
                "Uncertainty",
                "uncertainty_samples",
                self.util.get_exp_name(),
            )

    def set_id(self, run, epoch):
        """Make the report identifiable with run and epoch index."""
        self.run = run
        self.epoch = epoch

    def continuous_to_categorical(self):
        if self.cont_to_cat:
            return
        self.cont_to_cat = True
        bins = ast.literal_eval(glob_conf.config["DATA"]["bins"])
        self.truths = np.digitize(self.truths, bins) - 1
        self.preds = np.digitize(self.preds, bins) - 1

    def plot_confmatrix(self, plot_name, epoch=None):
        """Plot a confusionmatrix to the store.

        Args:
            plot_name (str): name for the image file.
            epoch (int, optional): Number of epoch. Defaults to None.
        """
        if not self.util.exp_is_classification():
            self._plot_scatter(
                self.truths, self.preds, plot_name.replace("cnf", "scatter"), epoch
            )
            self.continuous_to_categorical()
        self._plot_confmat(self.truths, self.preds, plot_name, epoch)

    def plot_per_speaker(self, result_df, plot_name, function):
        """Plot a confusion matrix with the mode category per speakers.

        If the function is mode and the values continuous, bin first

        Args:
            result_df: a pandas dataframe with columns: preds, truths and speaker.
            plot_name: name for the figure.
            function: either mode or mean.
        """
        if function == "mode" and not self.is_classification:
            truths, preds = result_df["truths"].values, result_df["preds"].values
            truths, preds = self.util._bin_distributions(truths, preds)
            result_df["truths"], result_df["preds"] = truths, preds
        speakers = result_df.speakers.unique()
        preds_speakers = np.zeros(0)
        truths_speakers = np.zeros(0)
        for s in speakers:
            s_df = result_df[result_df.speakers == s]
            s_truth = s_df.truths.iloc[0]
            s_pred = None
            if function == "mode":
                s_pred = s_df.preds.mode().iloc[-1]
            elif function == "mean":
                s_pred = s_df.preds.mean()
            else:
                self.util.error(f"unknown function {function}")
            preds_speakers = np.append(preds_speakers, s_pred)
            truths_speakers = np.append(truths_speakers, s_truth)
        test_result, upper, lower = self._get_test_result(
            result_df.truths.values, result_df.preds.values, self.metric
        )
        test_result = Result(test_result, None, None, None, self.METRIC)
        test_result.set_upper_lower(upper, lower)
        result_msg = f"Speaker combination result: {test_result.test_result_str()}"
        self.util.debug(result_msg)
        if function == "mean":
            truths_speakers, preds_speakers = self.util._bin_distributions(
                truths_speakers, preds_speakers
            )
        self._plot_confmat(
            truths_speakers,
            preds_speakers.astype("int"),
            plot_name,
            test_result=test_result,
        )

    def _plot_scatter(self, truths, preds, plot_name, epoch=None):
        # print(truths)
        # print(preds)
        if epoch is None:
            epoch = self.epoch
        fig_dir = self.util.get_path("fig_dir")
        pcc = pearsonr(self.truths, self.preds)[0]
        reg_res = self.result.test_result_str()
        fig = plt.figure()
        plt.scatter(truths, preds)
        plt.xlabel("truth")
        plt.ylabel("prediction")

        if epoch != 0:
            plt.title(f"Scatter plot: {reg_res}, PCC: {pcc:.3f}, Epoch: {epoch}")
        else:
            plt.title(f"Scatter plot: {reg_res}, PCC: {pcc:.3f}")

        img_path = f"{fig_dir}{plot_name}{self.filenameadd}.{self.format}"
        plt.savefig(img_path)
        self.util.debug(f"Saved scatter plot to {img_path}")
        fig.clear()
        plt.close(fig)
        plt.close()
        plt.clf()
        glob_conf.report.add_item(
            ReportItem(
                Header.HEADER_RESULTS,
                self.util.get_model_description(),
                "Result scatter plot",
                img_path,
            )
        )

    def _plot_confmat(self, truths, preds, plot_name, epoch=None, test_result=None):
        if epoch is None:
            epoch = self.epoch
        if test_result is None:
            test_result = self.result
        fig_dir = self.util.get_path("fig_dir")
        labels = glob_conf.labels
        fig = plt.figure()  # figsize=[5, 5]
        uar, (upper, lower) = evaluate_with_conf_int(
            self.preds,
            unweighted_average_recall,
            self.truths,
            num_bootstraps=1000,
            alpha=5,
        )
        acc = accuracy(truths, preds)
        cm = confusion_matrix(
            truths, preds, normalize=None
        )  # normalize must be one of {'true', 'pred', 'all', None}
        if cm.shape[0] != len(labels):
            self.util.error(
                f"mismatch between confmatrix dim ({cm.shape[0]}) and labels"
                f" length ({len(labels)}: {labels})"
            )

        try:
            disp = ConfusionMatrixDisplay(
                confusion_matrix=cm, display_labels=labels
            ).plot(cmap="Blues")
        except ValueError:
            disp = ConfusionMatrixDisplay(
                confusion_matrix=cm,
                display_labels=list(labels).remove("neutral"),
            ).plot(cmap="Blues")

        reg_res = ""
        if not self.is_classification:
            reg_res = f"{test_result.test_result_str()}"
            self.util.debug(
                f"Best result at epoch {epoch}: {test_result.test_result_str()}"
            )

        uar_str = self.util.to_3_digits_str(uar)
        acc_str = self.util.to_3_digits_str(acc)
        up_str = self.util.to_3_digits_str(upper)
        low_str = self.util.to_3_digits_str(lower)

        if epoch != 0:
            plt.title(
                f"Confusion Matrix, UAR: {uar_str} "
                + f"(+-{up_str}/{low_str}), {reg_res}, Epoch: {epoch}"
            )
        else:
            plt.title(
                f"Confusion Matrix, UAR: {uar_str} "
                + f"(+-{up_str}/{low_str}) {reg_res}"
            )
        img_path = f"{fig_dir}{plot_name}{self.filenameadd}.{self.format}"
        plt.savefig(img_path)
        self.util.debug(f"Saved confusion plot to {img_path}")
        fig.clear()
        plt.close(fig)
        plt.close()
        plt.clf()
        glob_conf.report.add_item(
            ReportItem(
                Header.HEADER_RESULTS,
                self.util.get_model_description(),
                "Confusion matrix",
                img_path,
            )
        )

        res_dir = self.util.get_path("res_dir")
        rpt = (
            f"Best score at epoch: {epoch}, UAR: {uar_str}"
            + f", (+-{up_str}/{low_str}), ACC: {acc_str}"
        )
        # print(rpt)
        self.util.debug(rpt)
        file_name = f"{res_dir}{self.util.get_exp_name()}{self.filenameadd}_conf.txt"
        with open(file_name, "w") as text_file:
            text_file.write(rpt)

    def set_filename_add(self, my_string):
        self.filenameadd = f"_{my_string}"

    def print_results(self, epoch=None):
        if epoch is None:
            epoch = self.epoch
        """Print all evaluation values to text file."""
        res_dir = self.util.get_path("res_dir")
        fig_dir = self.util.get_path("fig_dir")
        file_name = f"{res_dir}{self.util.get_exp_name()}_{epoch}{self.filenameadd}.txt"
        if self.util.exp_is_classification():
            labels = glob_conf.labels
            try:
                rpt = classification_report(
                    self.truths,
                    self.preds,
                    target_names=labels,
                    output_dict=True,
                )
                # print classifcation report in console
                self.util.debug(
                    f"\n {classification_report(self.truths, self.preds, target_names=labels, digits=4)}"
                )
            except ValueError as e:
                self.util.debug(
                    "Reporter: caught a ValueError when trying to get"
                    " classification_report: " + e
                )
                rpt = self.result.to_string()
            with open(file_name, "w") as text_file:
                c_ress = list(range(len(labels)))
                for i, l in enumerate(labels):
                    c_res = rpt[l]["f1-score"]
                    c_ress[i] = float(f"{c_res:.3f}")
                self.util.debug(f"labels: {labels}")
                f1_per_class = (
                    f"result per class (F1 score): {c_ress} from epoch: {epoch}"
                )
                # the following auc is buggy, preds should be probabilities
                # if len(np.unique(self.truths)) == 2:
                #     fpr, tpr, _ = roc_curve(self.truths, self.preds)
                #     auc_score = auc(fpr, tpr)
                #     plot_path = f"{fig_dir}{self.util.get_exp_name()}_{epoch}{self.filenameadd}_roc.{self.format}"
                #     plt.figure()
                #     display = RocCurveDisplay(
                #         fpr=fpr,
                #         tpr=tpr,
                #         roc_auc=auc_score,
                #         estimator_name=f"{self.model_type} estimator",
                #     )
                #     display.plot(ax=None)
                #     plt.savefig(plot_path)
                #     plt.close()
                #     self.util.debug(f"Saved ROC curve to {plot_path}")
                #     pauc_score = roc_auc_score(self.truths, self.preds, max_fpr=0.1)
                #     auc_pauc = f"auc: {auc_score:.3f}, pauc: {pauc_score:.3f} from epoch: {epoch}"
                #     self.util.debug(auc_pauc)
                self.util.debug(f1_per_class)
                rpt_str = f"{json.dumps(rpt)}\n{f1_per_class}"
                # rpt_str += f"\n{auc_auc}"
                text_file.write(rpt_str)
                glob_conf.report.add_item(
                    ReportItem(
                        Header.HEADER_RESULTS,
                        f"Classification result {self.util.get_model_description()}",
                        rpt_str,
                    )
                )

        else:  # regression
            result = self.result.test
            r2 = r2_score(self.truths, self.preds)
            pcc = pearsonr(self.truths, self.preds)[0]
            measure = self.util.config_val("MODEL", "measure", "mse")
            with open(file_name, "w") as text_file:
                text_file.write(
                    f"{measure}: {result:.3f}, r_2: {r2:.3f}, pcc {pcc:.3f}"
                )

    def make_conf_animation(self, out_name):
        import imageio

        fig_dir = self.util.get_path("fig_dir")
        filenames = glob.glob(fig_dir + f"{self.util.get_plot_name()}*_?_???_cnf.png")
        images = []
        for filename in filenames:
            images.append(imageio.imread(filename))
        fps = self.util.config_val("PLOT", "fps", "1")
        try:
            imageio.mimsave(fig_dir + out_name, images, fps=int(fps))
        except RuntimeError as e:
            self.util.error("error writing anim gif: " + e)

    def get_result(self):
        return self.result

    def plot_epoch_progression_finetuned(self, df):
        plot_name_suggest = self.util.get_exp_name()
        fig_dir = self.util.get_path("fig_dir")
        plot_name = (
            self.util.config_val("PLOT", "name", plot_name_suggest)
            + "_epoch_progression"
        )
        ax = df.plot()
        fig = ax.figure
        plt.xlabel("epochs")
        plt.ylabel(f"{self.METRIC}")
        plot_path = f"{fig_dir}{plot_name}.{self.format}"
        plt.savefig(plot_path)
        self.util.debug(f"plotted epoch progression to {plot_path}")
        plt.close(fig)
        # fig.clear()

    def plot_epoch_progression(self, reports, out_name):
        fig_dir = self.util.get_path("fig_dir")
        results, losses, train_results, losses_eval = [], [], [], []
        for r in reports:
            results.append(r.get_result().test)
            losses.append(r.get_result().loss)
            train_results.append(r.get_result().train)
            losses_eval.append(r.get_result().loss_eval)

        # do a plot per run
        # scale the losses so they fit on the picture
        losses, results, train_results, losses_eval = (
            np.asarray(losses),
            np.asarray(results),
            np.asarray(train_results),
            np.asarray(losses_eval),
        )

        if np.all((results > 1)):
            # scale down values
            results = results / 100.0
            train_results = train_results / 100.0
        # if np.all((losses < 1)):
        # scale up values
        plt.figure(dpi=200)
        plt.plot(train_results, "green", label="train set")
        plt.plot(results, "red", label="dev set")
        plt.plot(losses, "black", label="losses")
        plt.plot(losses_eval, "grey", label="losses_eval")
        plt.xlabel("epochs")
        plt.ylabel(f"{self.METRIC}")
        plt.legend()
        plt.savefig(f"{fig_dir}{out_name}.{self.format}")
        plt.close()



"""
feats_spectra.py

Inspired by code from Su Lei

"""

import os
import pathlib

import audeer
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torchaudio
import torchaudio.transforms as T
from PIL import Image, ImageOps
from tqdm import tqdm

import nkululeko.glob_conf as glob_conf
from nkululeko.constants import SAMPLING_RATE
from nkululeko.feat_extract.featureset import Featureset


class Spectraloader(Featureset):
    def __init__(self, name, data_df, feat_type):
        """Constructor setting the name"""
        super().__init__(name, data_df, feat_type)
        self.sampling_rate = SAMPLING_RATE
        self.num_bands = int(self.util.config_val("FEATS", "fft_nbands", "64"))
        self.win_dur = int(self.util.config_val("FEATS", "fft_win_dur", "25"))
        self.hop_dur = int(self.util.config_val("FEATS", "fft_hop_dur", "10"))

    def extract(self):
        """Extract the features or load them from disk if present."""
        store = self.util.get_path("store")
        store_format = self.util.config_val("FEATS", "store_format", "pkl")
        storage = f"{store}{self.name}.{store_format}"
        extract = self.util.config_val("FEATS", "needs_feature_extraction", False)
        no_reuse = eval(self.util.config_val("FEATS", "no_reuse", "False"))
        if extract or no_reuse or not os.path.isfile(storage):
            self.util.debug("extracting mel spectra, this might take a while...")
            image_store = audeer.mkdir(f"{store}{self.name}")
            images = []
            for idx, (file, start, end) in enumerate(
                tqdm(self.data_df.index.to_list())
            ):
                signal, sampling_rate = torchaudio.load(
                    file,
                    frame_offset=int(start.total_seconds() * 16000),
                    num_frames=int((end - start).total_seconds() * 16000),
                )
                assert sampling_rate == 16000, f"got {sampling_rate} instead of 16000"
                image = self._waveform2rgb(signal)
                outfile = f"{image_store}/{pathlib.Path(file).stem}_{idx}.jpg"
                image.save(outfile)
                images.append(outfile)
            self.df = pd.DataFrame(images, index=self.data_df.index)
            self.util.write_store(self.df, storage, store_format)
            try:
                glob_conf.config["DATA"]["needs_feature_extraction"] = "false"
            except KeyError:
                pass
        else:
            self.util.debug("reusing extracted spectrograms")
            self.df = self.util.get_store(storage, store_format)

    def _waveform2rgb(self, waveform, target_size=(256, 256)):
        # Transform to spectrogram
        spectrogram = T.MelSpectrogram(
            sample_rate=SAMPLING_RATE,
            n_mels=self.num_bands,
            hop_length=int(self.hop_dur * SAMPLING_RATE / 1000),
            win_length=int(self.win_dur * SAMPLING_RATE / 1000),
        )(waveform)
        melspec = T.AmplitudeToDB()(spectrogram)[0].numpy()
        melspec_norm = (melspec - np.min(melspec)) / (np.max(melspec) - np.min(melspec))

        # Map normalized Mel spectrogram to viridis colormap
        cmapped = plt.get_cmap("viridis")(melspec_norm)

        # Convert this colormap representation to a format suitable for creating a PIL Image
        image_array = (cmapped[:, :, :3] * 255).astype(np.uint8)
        image = Image.fromarray(image_array, mode="RGB")
        image = ImageOps.flip(image)

        # Resize to target size
        image = image.resize(target_size, Image.Resampling.LANCZOS)
        return image



""" feats_snr.py is to estimate snr.

SNR (signal to noise ratio) is extracted as acoustic features.
"""

import os

import audiofile
import pandas as pd
from tqdm import tqdm

import nkululeko.glob_conf as glob_conf
from nkululeko.autopredict.estimate_snr import SNREstimator
from nkululeko.feat_extract.featureset import Featureset


class SNRSet(Featureset):
    """Class to estimate snr"""

    def __init__(self, name, data_df):
        """Constructor."""

        super().__init__(name, data_df)

    def extract(self):
        """Estimate the features or load them from disk if present."""

        store = self.util.get_path("store")
        store_format = self.util.config_val("FEATS", "store_format", "pkl")
        storage = f"{store}{self.name}.{store_format}"
        extract = self.util.config_val("FEATS", "needs_feature_extraction", False)
        no_reuse = eval(self.util.config_val("FEATS", "no_reuse", "False"))
        if extract or no_reuse or not os.path.isfile(storage):
            self.util.debug("estimating SNR, this might take a while...")
            snr_series = pd.Series(index=self.data_df.index, dtype=object)
            for idx, (file, start, end) in enumerate(
                tqdm(self.data_df.index.to_list())
            ):
                signal, sampling_rate = audiofile.read(
                    file,
                    offset=start.total_seconds(),
                    duration=(end - start).total_seconds(),
                    always_2d=True,
                )
                snr = self.get_snr(signal[0], sampling_rate)
                snr_series[idx] = snr
            print("")
            self.df = pd.DataFrame(snr_series.values.tolist(), index=self.data_df.index)
            self.df.columns = ["snr"]
            self.util.write_store(self.df, storage, store_format)
            try:
                glob_conf.config["DATA"]["needs_feature_extraction"] = "false"
            except KeyError:
                pass
        else:
            self.util.debug("reusing estimated SNR values")
            self.df = self.util.get_store(storage, store_format)

    def get_snr(self, signal, sampling_rate):
        r"""Estimate SNR from raw audio signal.

        Args:
            signal: audio signal
            sampling_rate: sample rate
        Returns:
            snr: estimated signal to noise ratio
        """
        snr_estimator = SNREstimator(signal, sampling_rate)
        (
            estimated_snr,
            log_energies,
            energy_threshold_low,
            energy_threshold_high,
        ) = snr_estimator.estimate_snr()
        return estimated_snr

    def extract_sample(self, signal, sr):
        self.init_model()
        feats = self.get_snr(signal, sr)
        return feats



# script to MSP-IMPROV label (ID, EmotionCategory, A, V, D, N)
# and split into train, dev, test, provided for Nkululeko
# bagustris@yahoo.com, 2023-08-22


import argparse
import csv
import os

import pandas as pd

train_list = []
dev_list = []
test_list = []

def main(data_path):
    with open(data_path + 'Evaluation.txt') as f:
        for line in f:
            if line[:3] == 'UTD':
                label = line.split(';')
               # Allocate Session6 for test
                filename = 'MSP-'+ label[0][4:-4]
                wavfile = os.path.join(
                    data_path, 
                    'session' + str(label[0].split('-')[3][-1]),
                    label[0].split('-')[2],
                    label[0].split('-')[4],
                    filename + '.wav')
                # gender information
                g = label[0].split('-')[3][0]
                # speaker information
                s = label[0].split('-')[3]
                emo_cat = label[1].strip()
                a = float(label[2][3:])
                v = float(label[3][3:])
                d = float(label[4][3:])
                n = float(label[5][3:])
                data = [wavfile, g, s, emo_cat, v, a, d, n]
                if label[0].split('-')[3][-2:] == '06':
                    test_list.append(data)
                elif label[0].split('-')[3][-2:] == '05':
                    dev_list.append(data)
                else:
                    train_list.append(data)

    test_list_df = pd.DataFrame(test_list).fillna(3)
    test_list_df.to_csv('improv_train.csv', index=False, header=['file', 'gender', 'speaker', 'emotion', 'valence', 'arousal', 'dominance', 'naturalness'])

    dev_list_df = pd.DataFrame(dev_list).fillna(3)
    dev_list_df.to_csv('improv_dev.csv', index=False, header=['file', 'gender', 'speaker', 'emotion', 'valence', 'arousal', 'dominance', 'naturalness'])

    train_list_df = pd.DataFrame(train_list).fillna(3)
    train_list_df.to_csv('improv_test.csv', index=False, header=['file', 'gender', 'speaker', 'emotion', 'valence', 'arousal', 'dominance', 'naturalness'])


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_path', type=str, default='./MSP-IMPROV/', help='path to dataset')
    args = parser.parse_args()
    main(args.data_path)



# modelrunner.py

import pandas as pd

from nkululeko import glob_conf
from nkululeko.utils.util import Util


class Modelrunner:
    """Class to model one run."""

    def __init__(self, df_train, df_test, feats_train, feats_test, run):
        """Constructor setting up the dataframes.

        Args:
            df_train: train dataframe
            df_test: test dataframe
            feats_train: train features
            feats_train: test features
        """
        self.df_train, self.df_test, self.feats_train, self.feats_test = (
            df_train,
            df_test,
            feats_train,
            feats_test,
        )
        self.util = Util("modelrunner")
        self.run = run
        self.target = glob_conf.config["DATA"]["target"]
        # intialize a new model
        model_type = glob_conf.config["MODEL"]["type"]
        self._select_model(model_type)
        self.best_performance = 0
        self.best_epoch = 0

    def do_epochs(self):
        # initialze results
        reports = []
        plot_epochs = self.util.config_val("PLOT", "epochs", False)
        only_test = self.util.config_val("MODEL", "only_test", False)
        epoch_num = int(self.util.config_val("EXP", "epochs", 1))
        if not self.model.is_ann() and epoch_num > 1:
            self.util.warn(f"setting epoch num to 1 (was {epoch_num}) if model not ANN")
            epoch_num = 1
            glob_conf.config["EXP"]["epochs"] = "1"
        patience = self.util.config_val("MODEL", "patience", False)
        patience_counter = -1
        if self.util.high_is_good():
            highest = 0
        else:
            highest = 100000
        if self.model.model_type == "finetuned":
            # epochs are handled by Huggingface API
            self.model.train()
            report = self.model.predict()
            # todo: findout the best epoch -> no need
            # since load_best_model_at_end is given in training args
            epoch = epoch_num
            report.set_id(self.run, epoch)
            plot_name = self.util.get_plot_name() + f"_{self.run}_{epoch:03d}_cnf"
            reports.append(report)
            self.util.debug(
                f"run: {self.run} epoch: {epoch}: result: "
                f"{reports[-1].get_result().get_test_result()}"
            )
            if plot_epochs:
                self.util.debug(f"plotting conf matrix to {plot_name}")
                report.plot_confmatrix(plot_name, epoch)
        else:
            # for all epochs
            for epoch in range(epoch_num):
                if only_test:
                    self.model.load(self.run, epoch)
                    self.util.debug(f"reusing model: {self.model.store_path}")
                    self.model.reset_test(self.df_test, self.feats_test)
                else:
                    self.model.set_id(self.run, epoch)
                    self.model.train()
                report = self.model.predict()
                report.set_id(self.run, epoch)
                plot_name = self.util.get_plot_name() + f"_{self.run}_{epoch:03d}_cnf"
                reports.append(report)
                test_score_metric = report.get_result().get_test_result()
                self.util.debug(
                    f"run: {self.run} epoch: {epoch}: result: {test_score_metric}"
                )
                # print(f"performance: {performance.split(' ')[1]}")
                performance = float(test_score_metric.split(" ")[1])
                if performance > self.best_performance:
                    self.best_performance = performance
                    self.best_epoch = epoch
                if plot_epochs:
                    self.util.debug(f"plotting conf matrix to {plot_name}")
                    report.plot_confmatrix(plot_name, epoch)
                store_models = self.util.config_val("EXP", "save", False)
                plot_best_model = self.util.config_val("PLOT", "best_model", False)
                if (store_models or plot_best_model) and (
                    not only_test
                ):  # in any case the model needs to be stored to disk.
                    self.model.store()
                if patience:
                    patience = int(patience)
                    result = report.result.get_result()
                    if self.util.high_is_good():
                        if result > highest:
                            highest = result
                            patience_counter = 0
                        else:
                            patience_counter += 1
                    else:
                        if result < highest:
                            highest = result
                            patience_counter = 0
                        else:
                            patience_counter += 1
                    if patience_counter >= patience:
                        self.util.debug(
                            f"reached patience ({str(patience)}): early stopping"
                        )
                        break
        # After training, report the best performance and epoch
        best_report = reports[self.best_epoch]
        # self.util.debug(f"Best score at epoch: {self.best_epoch}, UAR: {self.best_performance}") # move to reporter below

        if not plot_epochs:
            # Do at least one confusion matrix plot
            self.util.debug(f"plotting confusion matrix to {plot_name}")
            best_report.plot_confmatrix(plot_name, self.best_epoch)
        return reports, epoch

    def _select_model(self, model_type):
        self._check_balancing()

        if model_type == "svm":
            from nkululeko.models.model_svm import SVM_model

            self.model = SVM_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "svr":
            from nkululeko.models.model_svr import SVR_model

            self.model = SVR_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "xgb":
            from nkululeko.models.model_xgb import XGB_model

            self.model = XGB_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "xgr":
            from nkululeko.models.model_xgr import XGR_model

            self.model = XGR_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "bayes":
            from nkululeko.models.model_bayes import Bayes_model

            self.model = Bayes_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "finetune":
            from nkululeko.models.model_tuned import TunedModel

            self.model = TunedModel(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "gmm":
            from nkululeko.models.model_gmm import GMM_model

            self.model = GMM_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "knn":
            from nkululeko.models.model_knn import KNN_model

            self.model = KNN_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "knn_reg":
            from nkululeko.models.model_knn_reg import KNN_reg_model

            self.model = KNN_reg_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "lin_reg":
            from nkululeko.models.model_lin_reg import Lin_reg_model

            self.model = Lin_reg_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "tree":
            from nkululeko.models.model_tree import Tree_model

            self.model = Tree_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "tree_reg":
            from nkululeko.models.model_tree_reg import Tree_reg_model

            self.model = Tree_reg_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "cnn":
            from nkululeko.models.model_cnn import CNNModel

            self.model = CNNModel(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "mlp":
            from nkululeko.models.model_mlp import MLPModel

            self.model = MLPModel(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        elif model_type == "mlp_reg":
            from nkululeko.models.model_mlp_regression import MLP_Reg_model

            self.model = MLP_Reg_model(
                self.df_train, self.df_test, self.feats_train, self.feats_test
            )
        else:
            self.util.error(f"unknown model type: '{model_type}'")
        if self.util.exp_is_classification() and not self.model.is_classifier:
            self.util.error(
                "Experiment type set to classification but model type is not a"
                " classifier"
            )
        return self.model

    def _check_balancing(self):
        balancing = self.util.config_val("FEATS", "balancing", False)
        if balancing:
            orig_size = self.feats_train.shape[0]
            self.util.debug(f"balancing the training features with: {balancing}")
            if balancing == "ros":
                from imblearn.over_sampling import RandomOverSampler

                sampler = RandomOverSampler(random_state=42)
                X_res, y_res = sampler.fit_resample(
                    self.feats_train, self.df_train[self.target]
                )
            elif balancing == "smote":
                from imblearn.over_sampling import SMOTE

                sampler = SMOTE(random_state=42)
                X_res, y_res = sampler.fit_resample(
                    self.feats_train, self.df_train[self.target]
                )
            elif balancing == "adasyn":
                from imblearn.over_sampling import ADASYN

                sampler = ADASYN(random_state=42)
                X_res, y_res = sampler.fit_resample(
                    self.feats_train, self.df_train[self.target]
                )
            else:
                self.util.error(
                    f"unknown balancing algorithm: {balancing} (should be [ros|smote|adasyn])"
                )

            self.feats_train = X_res
            self.df_train = pd.DataFrame({self.target: y_res}, index=X_res.index)
            self.util.debug(
                f"balanced with: {balancing}, new size: {X_res.shape[0]} (was {orig_size})"
            )
            le = glob_conf.label_encoder
            res = y_res.value_counts()
            resd = {}
            for i, e in enumerate(le.inverse_transform(res.index.values)):
                resd[e] = res.values[i]
            self.util.debug(f"{resd})")



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath('../..'))


# -- Project information -----------------------------------------------------

project = 'nkululeko'
copyright = '2023, Felix Burkhardt, Bagus Tris Atmaja'
author = 'Felix Burkhardt, Bagus Tris Atmaja'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    # 'nbsphinx',           # to embedd ipynb files
    'sphinx.ext.mathjax',  # to enable latex
    'sphinx_rtd_theme',
    'myst_parser',        # to enable markdown
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# suffix for source files
source_suffix = {
    '.rst': 'restructuredtext',
    '.txt': 'restructuredtext',
    '.md': 'markdown',
}

# headed anchors until h3
myst_heading_anchors = 3

# The master toctree document.
master_doc = 'index'


# html_theme_options = {

#     # Set the name of the project to appear in the navigation.
# 'nav_title': 'Nkululeko',

#     # Set the color and the accent color
#     'color_primary': 'blue',
#     'color_accent': 'light-blue',

#     # Set the repo location to get a badge with stats
#     'repo_url': 'https://github.com/felixbur/nkululeko',
#     'repo_name': 'Machine learning speaker characteristics',

#     # Visible levels of the global TOC; -1 means unlimited
#     'globaltoc_depth': 3,
#     # If False, expand all TOC entries
#     'globaltoc_collapse': True,
#     # If True, show hidden TOC entries
#     'globaltoc_includehidden': False,
#     #'logo_icon': "&#xe913;"
# }



# randomsplicer.py

"""
Code originally by Oliver Pauly

Based on an idea by Klaus Scherer

K. R. Scherer, “Randomized splicing: A note on a simple technique for masking speech content” 
Journal of Experimental Research in Personality, vol. 5, pp. 155–159, 1971.

Evaluated in:
F. Burkhardt, Anna Derington, Matthias Kahlau, Klaus Scherer, Florian Eyben and Björn Schuller: Masking Speech Contents by Random Splicing: is Emotional Expression Preserved?, Proc. ICASSP, 2023

"""

import os

import audeer
import audiofile as af
import pandas as pd
from tqdm import tqdm

import nkululeko.augmenting.randomsplicing as rsp
from nkululeko.utils.util import Util


class Randomsplicer:
    """
    augmenting the train split
    """

    def __init__(self, df):
        self.df = df
        self.util = Util("randomsplicer")

    def run(self, sample_selection):
        """
        random splice the selected samples and return a dataframe with new files index.
            adjustable parameters:
            * p_reverse: probability of some samples to be in reverse order (default: 0.3)
            * top_db: top db level for silence to be recognized (default: 12)
        """

        p_reverse = float(self.util.config_val("AUGMENT", "p_reverse", "0.3"))
        top_db = float(self.util.config_val("AUGMENT", "top_db", "12"))

        files = self.df.index.get_level_values(0).values
        store = self.util.get_path("store")
        filepath = f"{store}randomspliced/"
        audeer.mkdir(filepath)
        self.util.debug(
            f"random splicing {sample_selection} samples to {filepath}, "
            + f"p_reverse = {p_reverse}, top_db = {top_db}",
        )
        newpath = ""
        index_map = {}
        for i, f in enumerate(tqdm(files)):
            signal, sr = af.read(f)
            filename = os.path.basename(f)
            parent = os.path.dirname(f).split("/")[-1]
            sig_new = rsp.random_splicing(
                signal,
                sr,
                p_reverse=p_reverse,
                top_db=top_db,
            )
            newpath = f"{filepath}/{parent}/"
            audeer.mkdir(newpath)
            new_full_name = newpath + filename
            af.write(new_full_name, signal=sig_new, sampling_rate=sr)
            index_map[f] = new_full_name

        df_ret = self.df.copy()

        file_index = df_ret.index.to_series().map(lambda x: index_map[x[0]]).values
        # workaround because i just couldn't get this easier...
        arrays = [
            file_index,
            list(df_ret.index.get_level_values(1)),
            list(df_ret.index.get_level_values(2)),
        ]
        new_index = pd.MultiIndex.from_arrays(arrays, names=("file", "start", "end"))
        df_ret = df_ret.set_index(new_index)

        return df_ret
